from .camera import CameraGeometry
from .atmosphere import get_atmosphere_profile_functions
from .telescope import TelescopeDescription
from .optics import OpticsDescription
from .subarray import SubarrayDescription


__all__ = [
    'CameraGeometry',
    'get_atmosphere_profile_functions',
    'TelescopeDescription',
    'OpticsDescription',
    'SubarrayDescription',
]



"""
common pytest fixtures for tests in ctapipe
"""

import pytest
from ctapipe.utils import get_dataset_path
from ctapipe.io.eventseeker import EventSeeker
from ctapipe.io.hessioeventsource import HESSIOEventSource

@pytest.fixture(scope='module')
def test_event():
    """ an example event for algorithm testing"""
    filename = get_dataset_path('gamma_test.simtel.gz')

    with HESSIOEventSource(input_url=filename) as reader:
        seeker = EventSeeker(reader)
        event = seeker['409']

    yield event






import os

import numpy as np
from astropy.utils.decorators import deprecated
from astropy import units as u

from ctapipe import utils
from ctapipe.utils import linalg


@deprecated(0.1, "will be replaced with proper coord transform")
def pixel_position_to_direction(pix_x, pix_y, tel_phi, tel_theta, tel_foclen):
    """
    TODO replace with proper implementation
    calculates the direction vector of corresponding to a
    (x,y) position on the camera

    beta is the pixel's angular distance to the centre
    according to beta / tel_view = r / maxR
    alpha is the polar angle between the y-axis and the pixel
    to find the direction the pixel is looking at:

    - the pixel direction is set to the telescope direction
    - offset by beta towards up
    - rotated around the telescope direction by the angle alpha


    Parameters
    -----------
    pix_x, pix_y : ndarray
        lists of x and y positions on the camera
    tel_phi, tel_theta: astropy quantities
        two angles that describe the orientation of the telescope
    tel_foclen : astropy quantity
        focal length of the telescope

    Returns
    -------
    pix_dirs : ndarray
        shape (n,3) list of "direction vectors"
        corresponding to a position on the camera
    """

    pix_alpha = np.arctan2(pix_y, pix_x)

    pix_rho = (pix_x ** 2 + pix_y ** 2) ** .5

    pix_beta = pix_rho / tel_foclen * u.rad

    tel_dir = linalg.set_phi_theta(tel_phi, tel_theta)

    pix_dirs = []
    for a, b in zip(pix_alpha, pix_beta):
        pix_dir = linalg.set_phi_theta(tel_phi, tel_theta - b)

        pix_dir = linalg.rotate_around_axis(pix_dir, tel_dir, a)
        pix_dirs.append(pix_dir * u.dimless)

    return pix_dirs


def alt_to_theta(alt):
    """transforms altitude (angle from the horizon upwards) to theta (angle from z-axis)
    for simtel array coordinate systems
    """
    return (90 * u.deg - alt).to(alt.unit)


def az_to_phi(az):
    """transforms azimuth (angle from north towards east)
    to phi (angle from x-axis towards y-axis)
    for simtel array coordinate systems
    """
    return -az


def transform_pixel_position(x, y):
    """transforms the x and y coordinates on the camera plane so that they correspond to
    the view as if looked along the pointing direction of the telescope, i.e. y->up and
    x->right
    """
    return x, -y



import json
import os

import pytest
import yaml

from ctapipe.utils import datasets


def test_find_datasets():
    # find all datasets matching pattern
    r = datasets.find_all_matching_datasets("(.*)\.camgeom\.fits\.gz")
    assert len(r) > 3

    # get the full filename for a resrouces
    assert os.path.exists(datasets.get_dataset_path(r[0]))

    # try using a pattern
    r = datasets.find_all_matching_datasets("(.*)\.camgeom\.fits\.gz",
                                            regexp_group=1)
    assert not r[0].endswith("gz")


def test_datasets_in_custom_path(tmpdir_factory):
    """
    check that a dataset in a user-defined CTAPIPE_SVC_PATH is located
    """

    tmpdir1 = tmpdir_factory.mktemp('datasets1')
    tmpdir2 = tmpdir_factory.mktemp('datasets2')
    os.environ['CTAPIPE_SVC_PATH'] = ":".join([str(tmpdir1), str(tmpdir2)])

    # create a dummy dataset to search for:

    dataset_name = "test_dataset_1.txt"
    dataset_path = str(tmpdir1.join(dataset_name))

    with open(dataset_path, "w") as fp:
        fp.write("test test test")

    # try to find dummy dataset
    path = datasets.get_dataset_path(dataset_name)
    assert path == dataset_path

    with pytest.raises(FileNotFoundError):
        datasets.get_dataset_path("does_not_exist")

    # try using find_all_matching_datasets:

    ds = datasets.find_all_matching_datasets("test.*",
                                             searchpath=os.environ[
                                                 'CTAPIPE_SVC_PATH'])
    assert dataset_name in ds


def test_structured_datasets(tmpdir):
    basename = "test.yml"

    test_data = dict(x=[1, 2, 3, 4, 5], y='test_json')

    os.environ['CTAPIPE_SVC_PATH'] = ":".join([str(tmpdir)])

    with tmpdir.join("data_test.json").open(mode='w') as fp:
        json.dump(test_data, fp)

    data1 = datasets.get_structured_dataset('data_test')
    assert data1['x'] == [1, 2, 3, 4, 5]
    assert data1['y'] == 'test_json'
    tmpdir.join("data_test.json").remove()

    test_data['y'] = 'test_yaml'
    with tmpdir.join("data_test.yaml").open(mode='w') as fp:
        yaml.dump(test_data, fp)

    data1 = datasets.get_structured_dataset('data_test')
    assert data1['x'] == [1, 2, 3, 4, 5]
    assert data1['y'] == 'test_yaml'



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Visualization: Methods for displaying data 
"""

try:
    from .mpl_camera import CameraDisplay
    from .mpl_array import ArrayDisplay
except ImportError:
    pass


__all__ = ['CameraDisplay', 'ArrayDisplay']



import re
from abc import ABCMeta, abstractmethod
from collections import defaultdict

from ctapipe.core import Component, Container

__all__ = ['TableReader', 'TableWriter']


class TableWriter(Component, metaclass=ABCMeta):
    """
    Base class for writing  Container classes as rows of an output table,
    where each `Field` becomes a column. Subclasses of this implement
    specific output types.

    See Also
    --------
    - `ctapipe.io.HDF5TableWriter`
    """

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self._transforms = defaultdict(dict)
        self._exclusions = defaultdict(list)

    def __enter__(self):

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):

        self.close()

    def exclude(self, table_name, pattern):
        """
        Exclude any columns (Fields)  matching the pattern from being written

        Parameters
        ----------
        table_name: str
            name of table on which to apply the exclusion
        pattern: str
            regular expression string to match column name
        """
        self._exclusions[table_name].append(re.compile(pattern))

    def _is_column_excluded(self, table_name, col_name):
        for pattern in self._exclusions[table_name]:
            if pattern.match(col_name):
                return True
        return False

    def add_column_transform(self, table_name, col_name, transform):
        """
        Add a transformation function for a column. This function will be
        called on the value in the container before it is written to the
        output file.

        Parameters
        ----------
        table_name: str
            identifier of table being written
        col_name: str
            name of column in the table (or item in the Container)
        transform: callable
            function that take a value and returns a new one
        """
        self._transforms[table_name][col_name] = transform
        self.log.debug("Added transform: {}/{} -> {}".format(table_name,
                                                             col_name,
                                                             transform))

    @abstractmethod
    def write(self, table_name, containers):
        """
        Write the contents of the given container or containers to a table.
        The first call to write  will create a schema and initialize the table
        within the file.
        The shape of data within the container must not change between calls,
        since variable-length arrays are not supported.

        Parameters
        ----------
        table_name: str
            name of table to write to
        container: `ctapipe.core.Container`
            container to write
        """
        pass

    @abstractmethod
    def open(self, filename, **kwargs):
        """
        open an output file

        Parameters
        ----------
        filename: str
            output file name
        kwargs:
            any extra args to pass to the subclass open method
        """
        pass

    @abstractmethod
    def close(self):
        pass

    def _apply_col_transform(self, table_name, col_name, value):
        """
        apply value transform function if it exists for this column
        """
        if col_name in self._transforms[table_name]:
            tr = self._transforms[table_name][col_name]
            value = tr(value)
        return value


class TableReader(Component, metaclass=ABCMeta):
    """
    Base class for row-wise table readers. Generally methods that read a
    full table at once are preferred to this method, since they are faster,
    but this can be used to re-play a table row by row into a
    `ctapipe.core.Container` class (the opposite of TableWriter)
    """

    def __init__(self):
        super().__init__()
        self._cols_to_read = defaultdict(list)
        self._transforms = defaultdict(dict)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def add_column_transform(self, table_name, col_name, transform):
        """
        Add a transformation function for a column. This function will be
        called on the value in the container before it is written to the
        output file.

        Parameters
        ----------
        table_name: str
            identifier of table being written
        col_name: str
            name of column in the table (or item in the Container)
        transform: callable
            function that take a value and returns a new one
        """
        self._transforms[table_name][col_name] = transform
        self.log.debug("Added transform: {}/{} -> {}".format(table_name,
                                                             col_name,
                                                             transform))

    def _apply_col_transform(self, table_name, col_name, value):
        """
        apply value transform function if it exists for this column
        """
        if col_name in self._transforms[table_name]:
            tr = self._transforms[table_name][col_name]
            value = tr(value)
        return value

    @abstractmethod
    def read(self, table_name: str, container: Container):
        """
        Returns a generator that reads the next row from the table into the
        given container.  The generator returns the same container. Note that
        no containers are copied, the data are overwritten inside.

        Parameters
        ----------
        table_name: str
            name of table to read from
        container : ctapipe.core.Container
            Container instance to fill
        """
        pass

    @abstractmethod
    def open(self, filename, **kwargs):
        pass

    @abstractmethod
    def close(self):
        pass



import copy
import pytest
from ctapipe.io.targetioeventsource import TargetIOEventSource
from ctapipe.io.eventsourcefactory import EventSourceFactory
from ctapipe.io.eventseeker import EventSeeker
from ctapipe.utils import get_dataset_path
from ctapipe.calib.camera.calibrator import CameraCalibrator

pytest.importorskip("target_driver")
pytest.importorskip("target_io")
pytest.importorskip("target_calib")


def test_chec_r1():
    url = get_dataset_path("chec_r1.tio")
    source = TargetIOEventSource(input_url=url)
    event = source._get_event_by_index(0)
    assert(source._r0_samples is None)
    assert(source._r1_samples.shape[1] == 2048)
    assert(round(source._r1_samples[0, 0, 0]) == -274)
    assert(event.r0.tels_with_data == {0})
    assert(event.r0.tel[0].waveform is None)
    assert(event.r1.tel[0].waveform[0, 0, 0] == source._r1_samples[0, 0, 0])


def test_event_id():
    url = get_dataset_path("chec_r1.tio")
    source = TargetIOEventSource(input_url=url)
    event_id = 2
    source._get_event_by_id(event_id)
    assert(event_id == source._reader.fCurrentEventID)
    assert(round(source._r1_samples[0, 0, 0]) == -274)


def test_singlemodule_r0():
    url = get_dataset_path("targetmodule_r0.tio")
    source = TargetIOEventSource(input_url=url)
    event = source._get_event_by_index(0)
    assert(round(source._r0_samples[0, 0, 0]) == 600)
    assert(source._r1_samples is None)
    assert(event.r0.tels_with_data == {0})
    assert(event.r0.tel[0].waveform[0, 0, 0] == source._r0_samples[0, 0, 0])


def test_singlemodule_r1():
    url = get_dataset_path("targetmodule_r1.tio")
    source = TargetIOEventSource(input_url=url)
    event = source._get_event_by_index(0)
    assert(source._r0_samples is None)
    assert(source._r1_samples.shape[1] == 64)
    assert(round(source._r1_samples[0, 0, 0]) == 0)
    assert(event.r0.tels_with_data == {0})
    assert(event.r0.tel[0].waveform is None)
    assert(event.r1.tel[0].waveform[0, 0, 0] == source._r1_samples[0, 0, 0])


def test_compatible():
    dataset = get_dataset_path("chec_r1.tio")
    assert TargetIOEventSource.is_compatible(dataset)

    dataset = get_dataset_path("gamma_test.simtel.gz")
    assert not TargetIOEventSource.is_compatible(dataset)


def test_stream():
    dataset = get_dataset_path("chec_r1.tio")
    with TargetIOEventSource(input_url=dataset) as source:
        assert not source.is_stream


def test_loop():
    dataset = get_dataset_path("chec_r1.tio")
    with TargetIOEventSource(input_url=dataset) as source:
        count = 0
        for event in source:
            assert event.r0.tels_with_data == {0}
            assert event.count == count
            count += 1

        for event in source:
            # Check generator has restarted from beginning
            assert event.count == 0
            break


def test_that_event_is_not_modified_after_loop():
    dataset = get_dataset_path("chec_r1.tio")
    with TargetIOEventSource(input_url=dataset, max_events=2) as source:
        for event in source:
            last_event = copy.deepcopy(event)

        # now `event` should be identical with the deepcopy of itself from
        # inside the loop.
        # Unfortunately this does not work:
        #      assert last_event == event
        # So for the moment we just compare event ids
        assert event.r0.event_id == last_event.r0.event_id


def test_len():
    dataset = get_dataset_path("chec_r1.tio")
    with TargetIOEventSource(input_url=dataset) as source:
        count = 0
        for _ in source:
            count += 1
        assert count == len(source)

    with TargetIOEventSource(input_url=dataset, max_events=3) as reader:
        assert len(reader) == 3


def test_geom():
    dataset = get_dataset_path("chec_r1.tio")
    with TargetIOEventSource(input_url=dataset) as source:
        event = source._get_event_by_index(0)
        assert event.inst.subarray.tels[0].camera.pix_x.size == 2048

    dataset = get_dataset_path("targetmodule_r1.tio")
    with TargetIOEventSource(input_url=dataset) as source:
        event = source._get_event_by_index(0)
        assert event.inst.subarray.tels[0].camera.pix_x.size == 64


def test_eventsourcefactory():
    dataset = get_dataset_path("chec_r1.tio")
    source = EventSourceFactory.produce(input_url=dataset)
    assert source.__class__.__name__ == "TargetIOEventSource"
    assert source.input_url == dataset


def test_eventseeker():
    dataset = get_dataset_path("chec_r1.tio")
    with TargetIOEventSource(input_url=dataset) as source:
        seeker = EventSeeker(source)
        event = seeker[0]
        assert source._event_index == 0
        assert source._event_id == 2
        assert event.count == 0
        assert event.r1.event_id == 2
        assert (round(source._r1_samples[0, 0, 0]) == -274)

        event = seeker['2']
        assert source._event_index == 0
        assert source._event_id == 2
        assert event.count == 0
        assert event.r1.event_id == 2
        assert (round(source._r1_samples[0, 0, 0]) == -274)

        event = seeker[-1]
        assert event.count == len(seeker) - 1

    with TargetIOEventSource(input_url=dataset, max_events=3) as source:
        with pytest.raises(IndexError):
            seeker = EventSeeker(source)
            _ = seeker[5]


def test_pipeline():
    dataset = get_dataset_path("chec_r1.tio")
    reader = TargetIOEventSource(input_url=dataset, max_events=10)
    calibrator = CameraCalibrator(eventsource=reader)
    for event in reader:
        calibrator.calibrate(event)
        assert event.r0.tel.keys() == event.dl1.tel.keys()



from collections import OrderedDict
import numpy as np
from pytest import raises
from ctapipe.utils.CutFlow import CutFlow, UndefinedCut, PureCountingCut


def smaller2(x):
    return x < 2


def smaller3(x):
    return x < 3


def smaller5(x):
    return x < 5


def test_CutFlow():
    flow = CutFlow("TestFlow")
    # set_cut and add_cut a aliases
    flow.set_cut("smaller5", smaller5)
    flow.add_cut("smaller3", lambda x: x < 3)

    for i in range(2, 6):
        flow.count("noCuts")
        # .keep counts if the function returns True,
        # i.e. when we "keep" the event
        if flow.keep("smaller5", i):
            # .cut counts if the function returns False,
            # i.e. when we do NOT "cut" the event
            if flow.cut("smaller3", i):
                pass
            else:
                # do something else that could fail or be rejected
                try:
                    assert i == 3
                    flow.count("something")
                except:
                    pass


    t = flow(sort_column=1)
    assert np.all(t["selected Events"] == [4, 3, 2, 1])

    with raises(UndefinedCut):
        flow.cut("undefined", 5)

    with raises(PureCountingCut):
        flow.cut("noCuts")


def test_set_cuts_clear():
    flow = CutFlow("TestFlow")
    flow.set_cut("smaller5", smaller5)

    flow.set_cuts(OrderedDict([
        ("smaller3", smaller3),
        ("smaller2", smaller2)
    ]), clear=True)

    assert flow.cuts == OrderedDict([
        ("smaller3", [smaller3, 0]),
        ("smaller2", [smaller2, 0])
    ])


def test_set_cuts_no_clear():
    flow = CutFlow("TestFlow")
    flow.set_cut("smaller5", smaller5)

    flow.set_cuts(OrderedDict([
        ("smaller3", smaller3),
        ("smaller2", smaller2)
    ]), clear=False)


    assert flow.cuts == OrderedDict([
        ("smaller5", [smaller5, 0]),
        ("smaller3", [smaller3, 0]),
        ("smaller2", [smaller2, 0])
    ])


if __name__ == "__main__":
    test_CutFlow()
    test_set_cuts_clear()
    test_set_cuts_no_clear()



from copy import deepcopy

import numpy as np
from numpy.testing import assert_almost_equal

from ctapipe.image.charge_extractors import (FullIntegrator,
                                             SimpleIntegrator,
                                             GlobalPeakIntegrator,
                                             LocalPeakIntegrator,
                                             NeighbourPeakIntegrator,
                                             ChargeExtractorFactory,
                                             AverageWfPeakIntegrator)


def test_full_integration(test_event):
    telid = 11
    event = deepcopy(test_event)
    data = event.r0.tel[telid].waveform
    nsamples = data.shape[2]
    ped = event.mc.tel[telid].pedestal
    data_ped = data - np.atleast_3d(ped / nsamples)
    data_ped = np.array([data_ped[0], data_ped[0]])  # Test LG functionality

    integrator = FullIntegrator()
    integration, peakpos, window = integrator.extract_charge(data_ped)

    assert_almost_equal(integration[0][0], 149, 0)
    assert_almost_equal(integration[1][0], 149, 0)
    assert peakpos[0][0] == 0
    assert peakpos[1][0] == 0


def test_simple_integration(test_event):
    telid = 11
    event = deepcopy(test_event)
    data = event.r0.tel[telid].waveform
    nsamples = data.shape[2]
    ped = event.mc.tel[telid].pedestal
    data_ped = data - np.atleast_3d(ped / nsamples)
    data_ped = np.array([data_ped[0], data_ped[0]])  # Test LG functionality

    integrator = SimpleIntegrator()
    integration, peakpos, window = integrator.extract_charge(data_ped)

    assert_almost_equal(integration[0][0], 74, 0)
    assert_almost_equal(integration[1][0], 74, 0)
    assert peakpos[0][0] == 0
    assert peakpos[1][0] == 0


def test_global_peak_integration(test_event):
    telid = 11
    event = deepcopy(test_event)
    data = event.r0.tel[telid].waveform
    nsamples = data.shape[2]
    ped = event.mc.tel[telid].pedestal
    data_ped = data - np.atleast_3d(ped / nsamples)
    data_ped = np.array([data_ped[0], data_ped[0]])  # Test LG functionality

    integrator = GlobalPeakIntegrator()
    integration, peakpos, window = integrator.extract_charge(data_ped)

    assert_almost_equal(integration[0][0], 58, 0)
    assert_almost_equal(integration[1][0], 58, 0)
    assert peakpos[0][0] == 14
    assert peakpos[1][0] == 14


def test_local_peak_integration(test_event):
    telid = 11
    event = deepcopy(test_event)
    data = event.r0.tel[telid].waveform
    nsamples = data.shape[2]
    ped = event.mc.tel[telid].pedestal
    data_ped = data - np.atleast_3d(ped / nsamples)
    data_ped = np.array([data_ped[0], data_ped[0]])  # Test LG functionality

    integrator = LocalPeakIntegrator()
    integration, peakpos, window = integrator.extract_charge(data_ped)

    assert_almost_equal(integration[0][0], 76, 0)
    assert_almost_equal(integration[1][0], 76, 0)
    assert peakpos[0][0] == 13
    assert peakpos[1][0] == 13


def test_nb_peak_integration(test_event):
    telid = 11
    event = deepcopy(test_event)
    data = event.r0.tel[telid].waveform
    nsamples = data.shape[2]
    ped = event.mc.tel[telid].pedestal
    data_ped = data - np.atleast_3d(ped / nsamples)
    data_ped = np.array([data_ped[0], data_ped[0]])  # Test LG functionality
    geom = event.inst.subarray.tel[telid].camera
    nei = geom.neighbor_matrix_where

    integrator = NeighbourPeakIntegrator()
    integrator.neighbours = nei
    integration, peakpos, window = integrator.extract_charge(data_ped)

    assert_almost_equal(integration[0][0], -64, 0)
    assert_almost_equal(integration[1][0], -64, 0)
    assert peakpos[0][0] == 20
    assert peakpos[1][0] == 20


def test_averagewf_peak_integration(test_event):
    telid = 11
    event = deepcopy(test_event)
    data = event.r0.tel[telid].waveform
    nsamples = data.shape[2]
    ped = event.mc.tel[telid].pedestal
    data_ped = data - np.atleast_3d(ped / nsamples)
    data_ped = np.array([data_ped[0], data_ped[0]])  # Test LG functionality

    integrator = AverageWfPeakIntegrator()
    integration, peakpos, window = integrator.extract_charge(data_ped)

    assert_almost_equal(integration[0][0], 73, 0)
    assert_almost_equal(integration[1][0], 73, 0)
    assert peakpos[0][0] == 10
    assert peakpos[1][0] == 10


def test_charge_extractor_factory(test_event):
    extractor = ChargeExtractorFactory.produce(

        product='LocalPeakIntegrator'
    )

    telid = 11
    event = deepcopy(test_event)
    data = event.r0.tel[telid].waveform
    nsamples = data.shape[2]
    ped = event.mc.tel[telid].pedestal
    data_ped = data - np.atleast_3d(ped / nsamples)

    integration, peakpos, window = extractor.extract_charge(data_ped)

    assert_almost_equal(integration[0][0], 76, 0)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from .fitshistogram import Histogram
from .json2fits import json_to_fits
from .dynamic_class import dynamic_class_from_module
from .table_interpolator import TableInterpolator
from .unstructured_interpolator import UnstructuredInterpolator
from .datasets import (find_all_matching_datasets, get_table_dataset, get_dataset_path,
                       find_in_path, get_dataset)
from .CutFlow import CutFlow, PureCountingCut, UndefinedCut



import numpy as np
import pytest

from ctapipe.calib.camera.gainselection import ThresholdGainSelector
from ctapipe.calib.camera.gainselection import SimpleGainSelector
from ctapipe.calib.camera.gainselection import pick_gain_channel


def test_pick_gain_channel():
    threshold = 100
    good_hg_value = 35
    good_lg_value = 50

    dummy_waveforms = np.ones((2, 1000, 30)) * good_lg_value
    dummy_waveforms[1:] = good_hg_value  # high gains

    # set pixels  above 500's sample 13 to a high value (to trigger switch)
    dummy_waveforms[0, 500:, 13:15] = threshold + 10

    # First test the default mode of replacing the full waveform:
    # at the end, the final waveforms should be of shape (1000,30) and
    # pixels from 500 and above should have the low-gain value of good_hg_value

    new_waveforms, gain_mask = pick_gain_channel(
        waveforms=dummy_waveforms,
        threshold=threshold,
        select_by_sample=False
    )

    assert gain_mask.shape == (1000,)
    assert new_waveforms.shape == (1000, 30)
    assert (new_waveforms[500:] == good_hg_value).all()
    assert (new_waveforms[:500] == good_lg_value).all()

    # Next test the optional mode of replacing only the samples above threshold:
    # at the end, the final waveforms should be of shape (1000,30) and
    # pixels from 500 and above and with sample number 15:13 should have the
    # low-gain value of good_hg_value:

    new_waveforms, gain_mask = pick_gain_channel(
        waveforms=dummy_waveforms,
        threshold=threshold,
        select_by_sample=True
    )

    assert gain_mask.shape == new_waveforms.shape
    assert new_waveforms.shape == (1000, 30)
    assert (new_waveforms[500:, 13:15] == good_hg_value).all()
    assert (new_waveforms[500:, :13] == good_lg_value).all()
    assert (new_waveforms[500:, 15:] == good_lg_value).all()


def test_pick_gain_channel_bad_input():
    input_waveforms = np.arange(10).reshape(1, 10)
    waveforms, gain_mask = pick_gain_channel(input_waveforms, threshold=4)
    assert gain_mask is not None
    assert (waveforms == input_waveforms).all()


def test_threshold_gain_selector():
    selector = ThresholdGainSelector()
    print(selector)

    assert 'NectarCam' in selector.thresholds

    threshold = selector.thresholds['NectarCam']
    good_hg_value = 35
    good_lg_value = 50
    dummy_waveforms = np.ones((2, 1000, 30)) * good_lg_value
    dummy_waveforms[1:] = good_hg_value  #
    dummy_waveforms[0, 500:, 13:15] = threshold + 10

    new_waveforms, gain_mask = selector.select_gains("NectarCam",
                                                     dummy_waveforms)
    assert gain_mask.shape == (1000,)
    assert new_waveforms.shape == (1000, 30)
    assert (new_waveforms[500:] == good_hg_value).all()
    assert (new_waveforms[:500] == good_lg_value).all()

    selector.select_by_sample = True

    new_waveforms, gain_mask = selector.select_gains("NectarCam",
                                                     dummy_waveforms)

    assert new_waveforms.shape == (1000, 30)
    assert (new_waveforms[500:, 13:15] == good_hg_value).all()
    assert (new_waveforms[500:, :13] == good_lg_value).all()
    assert (new_waveforms[500:, 15:] == good_lg_value).all()
    assert gain_mask.shape == new_waveforms.shape

    # test some failures:
    # Camera that doesn't have a threshold:
    with pytest.raises(KeyError):
        selector.select_gains("NonExistantCamera", dummy_waveforms)

    # 3-gain channel input:
    with pytest.raises(ValueError):
        selector.select_gains("NectarCam", np.ones((3, 1000, 30)))

    # 1-gain channel input:
    wf0 = np.ones((1, 1000, 1))
    wf1, gm = selector.select_gains("ASTRICam", wf0)
    assert wf1.shape == (1000,)
    assert gm.shape == (1000,)


def test_simple_gain_selector():
    gs = SimpleGainSelector()

    for chan in [0, 1]:
        gs.channel = chan

        waveforms_2g = np.random.normal(size=(2, 1000, 30))
        waveforms_1g, mask = gs.select_gains("NectarCam", waveforms_2g)

        assert waveforms_1g.shape == (1000, 30)
        assert (waveforms_1g == waveforms_2g[chan]).all()
        assert mask.shape == (1000,)



import pytest
from sklearn.ensemble import RandomForestClassifier

from ctapipe.reco.regressor_classifier_base import *


def test_reshuffle_event_list():
    feature_list = [
        {"FlashCam": [[1, 10], [2, 20]],
         "ASTRICam": [[30, 3], [40, 4]]},
        {"FlashCam": [[1.5, 15]],
         "ASTRICam": [[35, 3.5], [25, 2.5], [30, 3]]}
    ]
    target_list = ["1", "2"]

    cam_id_list = ["FlashCam", "ASTRICam"]
    my_base = RegressorClassifierBase(model=RandomForestClassifier,
                                      cam_id_list=cam_id_list, unit=1)

    feature_flattened, target_flattened = my_base.reshuffle_event_list(
        feature_list, target_list
    )

    assert feature_flattened == {'FlashCam': [[1, 10], [2, 20], [1.5, 15]],
                                 'ASTRICam': [[30, 3], [40, 4], [35, 3.5],
                                              [25, 2.5],
                                              [30, 3]]}

    assert target_flattened == {'FlashCam': ['1', '1', '2'],
                                'ASTRICam': ['1', '1', '2', '2', '2']}

    assert len(str(my_base)) > 0


def test_failures():
    cam_id_list = ["FlashCam", "ASTRICam"]
    my_base = RegressorClassifierBase(model=RandomForestClassifier,
                                      cam_id_list=cam_id_list, unit=1)

    # some test data with a bad camera in it should raise a KeyError
    feature_list = [
        {"FlashCam": [[1, 10], [2, 20]],
         "BadCam": [[30, 3], [40, 4]]},
        {"FlashCam": [[1.5, 15]],
         "ASTRICam": [[35, 3.5], [25, 2.5], [30, 3]]}
    ]
    target_list = ["1", "2"]

    with pytest.raises(KeyError):
        feature_flattened, target_flattened = my_base.reshuffle_event_list(
            feature_list,
            target_list
        )
        assert feature_flattened is not None
        assert target_flattened is not None



from ctapipe.instrument import CameraGeometry
from ctapipe.image import tailcuts_clean, toymodel
from ctapipe.image.hillas import (hillas_parameters_1, hillas_parameters_2,
                                  hillas_parameters_3, hillas_parameters_4,
                                  HillasParameterizationError)
from ctapipe.io.containers import HillasParametersContainer
from astropy import units as u
from numpy import isclose, zeros_like, arange
from numpy.random import seed
from numpy.ma import masked_array
import pytest


def create_sample_image(psi='-30d'):

    seed(10)

    # set up the sample image using a HESS camera geometry (since it's easy
    # to load)
    geom = CameraGeometry.from_name("LSTCam")

    # make a toymodel shower model
    model = toymodel.generate_2d_shower_model(centroid=(0.2, 0.3),
                                              width=0.001, length=0.01,
                                              psi=psi)

    # generate toymodel image in camera for this shower model.
    image, signal, noise = toymodel.make_toymodel_shower_image(geom, model.pdf,
                                                               intensity=50,
                                                               nsb_level_pe=100)

    # denoise the image, so we can calculate hillas params
    clean_mask = tailcuts_clean(geom, image, 10,
                                5)  # pedvars = 1 and core and boundary

    return geom, image, clean_mask


def create_sample_image_zeros(psi='-30d'):

    geom, image, clean_mask = create_sample_image(psi)

    # threshold in pe
    image[~clean_mask] = 0

    return geom, image


def create_sample_image_masked(psi='-30d'):
    geom, image, clean_mask = create_sample_image(psi)

    # threshold in pe
    image = masked_array(image, mask=~clean_mask)

    return geom, image


def compare_result(x, y):
    ux = u.Quantity(x)
    uy = u.Quantity(y)
    assert isclose(ux.value, uy.value)
    assert ux.unit == uy.unit


def test_hillas():
    """
    test all Hillas-parameter routines on a sample image and see if they
    agree with eachother and with the toy model (assuming the toy model code
    is correct)
    """

    # try all quadrants
    for psi_angle in ['30d', '120d', '-30d', '-120d']:

        geom, image = create_sample_image_zeros(psi_angle)
        results = {}

        results['v1'] = hillas_parameters_1(geom, image)
        results['v2'] = hillas_parameters_2(geom, image)
        results['v3'] = hillas_parameters_3(geom, image)
        results['v4'] = hillas_parameters_4(geom, image)
        # compare each method's output
        for aa in results:
            for bb in results:
                if aa is not bb:
                    print("comparing {} to {}".format(aa, bb))
                    compare_result(results[aa].length, results[bb].length)
                    compare_result(results[aa].width, results[bb].width)
                    compare_result(results[aa].r, results[bb].r)
                    compare_result(results[aa].phi.deg, results[bb].phi.deg)
                    compare_result(results[aa].psi.deg, results[bb].psi.deg)
                    compare_result(results[aa].miss, results[bb].miss)
                    compare_result(results[aa].skewness, results[bb].skewness)
                    # compare_result(results[aa].kurtosis, results[bb].kurtosis)


def test_hillas_masked():
    """
    test Hillas-parameter routines on a sample image with masked values set to
    zero against a sample image with values masked with a numpy.ma.masked_array
    """

    geom, image = create_sample_image_zeros()
    geom, image_ma = create_sample_image_masked()

    results = hillas_parameters_4(geom, image)
    results_ma = hillas_parameters_4(geom, image_ma)

    compare_result(results.length, results_ma.length)
    compare_result(results.width, results_ma.width)
    compare_result(results.r, results_ma.r)
    compare_result(results.phi.deg, results_ma.phi.deg)
    compare_result(results.psi.deg, results_ma.psi.deg)
    compare_result(results.miss, results_ma.miss)
    compare_result(results.skewness, results_ma.skewness)
    # compare_result(results.kurtosis, results_ma.kurtosis)


def test_hillas_failure():
    geom, image = create_sample_image_zeros(psi='0d')
    blank_image = zeros_like(image)

    with pytest.raises(HillasParameterizationError):
        hillas_parameters_1(geom, blank_image)

    with pytest.raises(HillasParameterizationError):
        hillas_parameters_2(geom, blank_image)

    with pytest.raises(HillasParameterizationError):
        hillas_parameters_3(geom, blank_image)

    with pytest.raises(HillasParameterizationError):
        hillas_parameters_4(geom, blank_image)


def test_hillas_api_change():
    import numpy as np
    with pytest.raises(ValueError):
        hillas_parameters_4(arange(10), arange(10), arange(10))


def test_hillas_container():
    geom, image = create_sample_image_zeros(psi='0d')
    params = hillas_parameters_4(geom, image, container=True)
    assert type(params) is HillasParametersContainer



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import os
import sys
import logging
from importlib import import_module
from ctapipe.core.tool import Tool



__all__ = ['dynamic_class_from_module', ]


class DynamicClassError(Exception):

    def __init__(self, msg):
        """Mentions that an exception occurred in the dynamic_class_from_module.
        """
        self.msg = msg

# def dynamic_class_from_module(class_name, module,  configuration=None):


def dynamic_class_from_module(class_name, module, tool=None):
    """
    Create an instance of a class from a configuration service section name

    Parameters
    ----------
    module: str
        a python module file name. This module containe class to instantiate
    class_name: str
        python class name contained in module

    Returns
    -------
    A python object instance of a class_name

    Raises
    ------
    """
    if module == None:
        return None

    try:
        _class = getattr(import_module(module), class_name)
        if isinstance(tool, Tool):
            instance = _class(tool, config=tool.config)
        else:
            instance = _class()
        return instance
    except AttributeError as e:
        raise DynamicClassError("Could not create an instance of {} in module {}: {}"
                                .format(class_name, module, e))
    except ImportError as e:
        raise DynamicClassError("Could not create an instance of {} in module {}: {}"
                                .format(class_name, module, e))
    except TypeError as e:
        raise DynamicClassError("Could not create an instance of {} in module {}: {}"
                                .format(class_name, module, e))



import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord

from ctapipe.instrument import (
    SubarrayDescription,
    TelescopeDescription,
)


def test_subarray_description():
    pos = {}
    tel = {}
    n_tels = 10

    for ii in range(n_tels):
        tel[ii + 1] = TelescopeDescription.from_name(optics_name="MST",
                                                     camera_name="NectarCam")
        pos[ii + 1] = np.random.uniform(-100, 100, size=3) * u.m

    sub = SubarrayDescription("test array",
                              tel_positions=pos,
                              tel_descriptions=tel)

    sub.info()

    assert str(sub) == "test array"
    assert sub.num_tels == n_tels
    assert len(sub.tel_ids) == n_tels
    assert sub.tel_ids[0] == 1
    assert sub.tel[1].camera is not None
    assert 0 not in sub.tel  # check that there is no tel 0 (1 is first above)
    assert len(sub.to_table()) == n_tels
    assert len(set(sub.camera_types)) == 1  # only 1 camera type
    assert sub.camera_types[0] == 'NectarCam'
    assert sub.optics_types[0] == 'MST'
    assert sub.telescope_types[0] == 'MST:NectarCam'
    assert sub.tel_coords
    assert isinstance(sub.tel_coords, SkyCoord)
    assert len(sub.tel_coords) == n_tels

    subsub = sub.select_subarray("newsub", [2, 3, 4, 6])
    assert subsub.num_tels == 4
    assert set(subsub.tels.keys()) == {2, 3, 4, 6}
    assert subsub.tel_indices[6] == 3
    assert subsub.tel_ids[3] == 6






import pytest
from traitlets import TraitError

from ctapipe.io.eventsourcefactory import EventSourceFactory, event_source
from ctapipe.utils import get_dataset_path


def test_factory_subclasses():
    factory = EventSourceFactory()
    assert len(factory.subclass_names) > 0


def test_factory():
    dataset = get_dataset_path("gamma_test.simtel.gz")
    reader = EventSourceFactory.produce(input_url=dataset)
    assert reader.__class__.__name__ == "HESSIOEventSource"
    assert reader.input_url == dataset


def test_factory_different_file():
    dataset = get_dataset_path("gamma_test_large.simtel.gz")
    reader = EventSourceFactory.produce(input_url=dataset)
    assert reader.__class__.__name__ == "HESSIOEventSource"
    assert reader.input_url == dataset


def test_factory_from_reader():
    dataset = get_dataset_path("gamma_test.simtel.gz")
    reader = EventSourceFactory.produce(
        product='HESSIOEventSource',
        input_url=dataset
    )
    assert reader.__class__.__name__ == "HESSIOEventSource"
    assert reader.input_url == dataset


def test_factory_unknown_file_format():
    with pytest.raises(ValueError):
        dataset = get_dataset_path("optics.ecsv.txt")
        reader = EventSourceFactory.produce(input_url=dataset)
        assert reader is not None


def test_factory_unknown_reader():
    with pytest.raises(TraitError):
        dataset = get_dataset_path("gamma_test.simtel.gz")
        reader = EventSourceFactory.produce(
            product='UnknownFileReader',
            input_url=dataset
        )
        assert reader is not None

def test_factory_incompatible_file():
    dataset = get_dataset_path("optics.ecsv.txt")
    reader = EventSourceFactory.produce(
        product='HESSIOEventSource',
        input_url=dataset
    )
    event_list = [event for event in reader]
    assert len(event_list) == 0
    # TODO: Need better test for this, why does pyhessio not throw an error?


def test_factory_nonexistant_file():
    with pytest.raises(FileNotFoundError):
        dataset = "/fake_path/fake_file.fake_extension"
        reader = EventSourceFactory.produce(
            product='HESSIOEventSource',
            input_url=dataset
        )
        assert reader is not None

def test_factory_incorrect_use():
    with pytest.raises(FileNotFoundError):
        dataset = get_dataset_path("gamma_test_large.simtel.gz")
        factory = EventSourceFactory(input_url=dataset)
        reader = factory.produce()
        assert reader is not None

def test_event_source_helper():
    with event_source(get_dataset_path("gamma_test_large.simtel.gz")) as source:
        for _ in source:
            pass



import tempfile

import numpy as np
import pytest
import tables
from astropy import units as u

from ctapipe.core.container import Container, Field
from ctapipe.io.containers import R0CameraContainer, MCEventContainer
from ctapipe.io.hdf5tableio import HDF5TableWriter, HDF5TableReader


@pytest.fixture(scope='session')
def temp_h5_file(tmpdir_factory):
    """ a fixture that fetches a temporary output dir/file for a test
    file that we want to read or write (so it doesn't clutter up the test
    directory when the automated tests are run)"""
    return str(tmpdir_factory.mktemp('data').join('test.h5'))


def test_write_container(temp_h5_file):
    r0tel = R0CameraContainer()
    mc = MCEventContainer()
    mc.reset()
    r0tel.waveform = np.random.uniform(size=(50, 10))
    r0tel.image = np.random.uniform(size=50)
    r0tel.num_samples = 10
    r0tel.meta['test_attribute'] = 3.14159
    r0tel.meta['date'] = "2020-10-10"

    writer = HDF5TableWriter(str(temp_h5_file), group_name='R0',
                             filters=tables.Filters(
        complevel=7))
    writer.exclude("tel_002", ".*samples")  # test exclusion of columns

    for ii in range(100):
        r0tel.waveform[:] = np.random.uniform(size=(50, 10))
        r0tel.image[:] = np.random.uniform(size=50)
        r0tel.num_samples = 10
        mc.energy = 10**np.random.uniform(1, 2) * u.TeV
        mc.core_x = np.random.uniform(-1, 1) * u.m
        mc.core_y = np.random.uniform(-1, 1) * u.m

        writer.write("tel_001", r0tel)
        writer.write("tel_002", r0tel)  # write a second table too
        writer.write("MC", mc)

    writer.close()


def test_write_containers(temp_h5_file):

    class C1(Container):
        a = Field('a', None)
        b = Field('b', None)

    class C2(Container):
        c = Field('c', None)
        d = Field('d', None)

    with tempfile.NamedTemporaryFile() as f:
        writer = HDF5TableWriter(f.name, 'test')
        for i in range(20):
            c1 = C1()
            c2 = C2()
            c1.a, c1.b, c2.c, c2.d = np.random.normal(size=4)
            c1.b = np.random.normal()

            writer.write("tel_001", [c1, c2])

        writer.close()


def test_read_container(temp_h5_file):
    r0tel1 = R0CameraContainer()
    r0tel2 = R0CameraContainer()
    mc = MCEventContainer()

    reader = HDF5TableReader(str(temp_h5_file))

    # get the generators for each table
    mctab = reader.read('/R0/MC', mc)
    r0tab1 = reader.read('/R0/tel_001', r0tel1)
    r0tab2 = reader.read('/R0/tel_002', r0tel2)

    # read all 3 tables in sync
    for ii in range(3):

        m = next(mctab)
        r0_1 = next(r0tab1)
        r0_2 = next(r0tab2)

        print("MC:", m)
        print("t0:", r0_1.image)
        print("t1:", r0_2.image)
        print("---------------------------")

    assert 'test_attribute' in r0_1.meta
    assert r0_1.meta['date'] == "2020-10-10"

    reader.close()


def test_read_whole_table(temp_h5_file):

    mc = MCEventContainer()

    reader = HDF5TableReader(str(temp_h5_file))

    for cont in reader.read('/R0/MC', mc):
        print(cont)

    reader.close()


def test_with_context_writer(temp_h5_file):

    class C1(Container):
        a = Field('a', None)
        b = Field('b', None)

    with tempfile.NamedTemporaryFile() as f:

        with HDF5TableWriter(f.name, 'test') as h5_table:

            for i in range(5):
                c1 = C1()
                c1.a, c1.b = np.random.normal(size=2)

                h5_table.write("tel_001", c1)


def test_writer_closes_file(temp_h5_file):

    with tempfile.NamedTemporaryFile() as f:
        with HDF5TableWriter(f.name, 'test') as h5_table:

            assert h5_table._h5file.isopen == True

    assert h5_table._h5file.isopen == False


def test_reader_closes_file(temp_h5_file):

    with HDF5TableReader(str(temp_h5_file)) as h5_table:

        assert h5_table._h5file.isopen == True

    assert h5_table._h5file.isopen == False


def test_with_context_reader(temp_h5_file):

    mc = MCEventContainer()

    with HDF5TableReader(str(temp_h5_file)) as h5_table:

        assert h5_table._h5file.isopen == True

        for cont in h5_table.read('/R0/MC', mc):
            print(cont)

    assert h5_table._h5file.isopen == False


def test_closing_reader(temp_h5_file):

    f = HDF5TableReader(str(temp_h5_file))
    f.close()


def test_closing_writer(temp_h5_file):

    with tempfile.NamedTemporaryFile() as f:
        h5_table = HDF5TableWriter(f.name, 'test')
        h5_table.close()


def test_cannot_read_with_writer(temp_h5_file):

    with pytest.raises(IOError):

        with HDF5TableWriter(temp_h5_file, 'test', mode='r'):

            pass


def test_cannot_write_with_reader(temp_h5_file):

    with HDF5TableReader(temp_h5_file, mode='w') as h5:

        assert h5._h5file.mode == 'r'


def test_cannot_append_with_reader(temp_h5_file):

    with HDF5TableReader(temp_h5_file, mode='a') as h5:

        assert h5._h5file.mode == 'r'


def test_cannot_r_plus_with_reader(temp_h5_file):

    with HDF5TableReader(temp_h5_file, mode='r+') as h5:

        assert h5._h5file.mode == 'r'


def test_append_mode(temp_h5_file):

    class ContainerA(Container):

        a = Field(int)

    a = ContainerA()
    a.a = 1

    # First open with 'w' mode to clear the file and add a Container
    with HDF5TableWriter(temp_h5_file, 'group') as h5:

        h5.write('table_1', a)

    # Try to append A again
    with HDF5TableWriter(temp_h5_file, 'group', mode='a') as h5:

        h5.write('table_2', a)

    # Check if file has two tables with a = 1
    with HDF5TableReader(temp_h5_file) as h5:

        for a in h5.read('/group/table_1', ContainerA()):

            assert a.a == 1

        for a in h5.read('/group/table_2', ContainerA()):

            assert a.a == 1


@pytest.mark.xfail
def test_write_to_any_location(temp_h5_file):

    loc = '/path/path_1'

    class ContainerA(Container):

        a = Field(int)

    a = ContainerA()
    a.a = 1

    with HDF5TableWriter(temp_h5_file, 'group_1', root_uep=loc) as h5:

        for _ in range(5):

            h5.write('table', a)

    with HDF5TableReader(temp_h5_file) as h5:

        for a in h5.read(loc + '/group_1/table', ContainerA()):

            assert a.a == 1


if __name__ == '__main__':

    import logging
    logging.basicConfig(level=logging.DEBUG)

    test_write_container("test.h5")
    test_read_container("test.h5")
    test_read_whole_table("test.h5")



from ctapipe.utils import get_dataset_path
from ctapipe.io.eventsource import EventSource


def test_construct():
    try:
        EventSource(config=None, tool=None)
    except TypeError:
        return
    raise TypeError("EventSource should raise a TypeError when "
                    "instantiated due to its abstract methods")


class DummyReader(EventSource):
    """
    Simple working EventSource
    """

    def _generator(self):
        return range(len(self.input_url))

    @staticmethod
    def is_compatible(file_path):
        return False


def test_can_be_implemented():
    dataset = get_dataset_path("gamma_test.simtel.gz")
    test_reader = DummyReader(input_url=dataset)
    assert test_reader is not None


def test_is_iterable():
    dataset = get_dataset_path("gamma_test.simtel.gz")
    test_reader = DummyReader(input_url=dataset)
    for _ in test_reader:
        pass



"""
Extract data necessary to calcualte charge resolution from raw data files.
"""

import os

import numpy as np
from tqdm import tqdm
from traitlets import Dict, List, Int, Unicode

from ctapipe.analysis.camera.chargeresolution import ChargeResolutionCalculator
from ctapipe.calib.camera.dl0 import CameraDL0Reducer
from ctapipe.calib.camera.dl1 import CameraDL1Calibrator
from ctapipe.calib.camera.r1 import HESSIOR1Calibrator
from ctapipe.core import Tool
from ctapipe.image.charge_extractors import ChargeExtractorFactory
from ctapipe.io.hessioeventsource import HESSIOEventSource


class ChargeResolutionGenerator(Tool):
    name = "ChargeResolutionGenerator"
    description = "Generate the a pickle file of ChargeResolutionFile for " \
                  "a MC file."

    telescopes = List(Int, None, allow_none=True,
                      help='Telescopes to include from the event file. '
                           'Default = All telescopes').tag(config=True)
    output_name = Unicode('charge_resolution',
                          help='Name of the output charge resolution hdf5 '
                               'file').tag(config=True)

    aliases = Dict(dict(f='HESSIOEventSource.input_url',
                        max_events='HESSIOEventSource.max_events',
                        extractor='ChargeExtractorFactory.product',
                        window_width='ChargeExtractorFactory.window_width',
                        t0='ChargeExtractorFactory.t0',
                        window_shift='ChargeExtractorFactory.window_shift',
                        sig_amp_cut_HG='ChargeExtractorFactory.sig_amp_cut_HG',
                        sig_amp_cut_LG='ChargeExtractorFactory.sig_amp_cut_LG',
                        lwt='ChargeExtractorFactory.lwt',
                        clip_amplitude='CameraDL1Calibrator.clip_amplitude',
                        radius='CameraDL1Calibrator.radius',
                        max_pe='ChargeResolutionCalculator.max_pe',
                        T='ChargeResolutionGenerator.telescopes',
                        O='ChargeResolutionGenerator.output_name',
                        ))
    classes = List([HESSIOEventSource,
                    ChargeExtractorFactory,
                    CameraDL1Calibrator,
                    ChargeResolutionCalculator
                    ])

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.eventsource = None
        self.r1 = None
        self.dl0 = None
        self.dl1 = None
        self.calculator = None

    def setup(self):
        self.log_format = "%(levelname)s: %(message)s [%(name)s.%(funcName)s]"
        kwargs = dict(config=self.config, tool=self)

        self.eventsource = HESSIOEventSource(**kwargs)

        extractor = ChargeExtractorFactory.produce(**kwargs)

        self.r1 = HESSIOR1Calibrator(**kwargs)

        self.dl0 = CameraDL0Reducer(**kwargs)

        self.dl1 = CameraDL1Calibrator(extractor=extractor, **kwargs)

        self.calculator = ChargeResolutionCalculator(**kwargs)

    def start(self):
        desc = "Filling Charge Resolution"
        for event in tqdm(self.eventsource, desc=desc):
            tels = list(event.dl0.tels_with_data)

            # Check events have true charge included
            if event.count == 0:
                try:
                    if np.all(event.mc.tel[
                                  tels[0]].photo_electron_image == 0):
                        raise KeyError
                except KeyError:
                    self.log.exception('Source does not contain '
                                       'true charge!')
                    raise

            self.r1.calibrate(event)
            self.dl0.reduce(event)
            self.dl1.calibrate(event)

            if self.telescopes:
                tels = []
                for tel in self.telescopes:
                    if tel in event.dl0.tels_with_data:
                        tels.append(tel)

            for telid in tels:
                true_charge = event.mc.tel[telid].photo_electron_image
                measured_charge = event.dl1.tel[telid].image[0]
                self.calculator.add_charges(true_charge, measured_charge)

    def finish(self):
        input_url = self.eventsource.input_url
        input_directory = os.path.dirname(input_url)
        input_name = os.path.splitext(os.path.basename(input_url))[0]
        output_directory = os.path.join(input_directory, input_name)
        if not os.path.exists(output_directory):
            self.log.info("Creating directory: {}".format(output_directory))
            os.makedirs(output_directory)
        name = "{}.h5".format(self.output_name)
        ouput_path = os.path.join(output_directory, name)
        self.calculator.save(ouput_path)


def main():
    exe = ChargeResolutionGenerator()
    exe.run()


if __name__ == '__main__':
    main()



from ctapipe.utils.unstructured_interpolator import UnstructuredInterpolator
import numpy as np
from scipy.interpolate import LinearNDInterpolator, RegularGridInterpolator


def test_simple_interpolation():
    """
    Simple test function to asses the basic funtionality of the unstructured
    interpolator, check if we just spread points on a grid
    """

    interpolation_points = {(0, 0): 0.,
                            (0, 1): 0.,
                            (1, 0): 1.,
                            (1, 1): 1.}

    interpolator = UnstructuredInterpolator(interpolation_points)

    # OK lets first just check we get the values at the grid points back out again...
    interpolated_point = interpolator([[0, 0], [0, 1],
                                       [1, 0], [1, 1]])
    assert np.all(interpolated_point == [0., 0., 1., 1.])

    interpolated_point = interpolator([[0., 0.5], [0.5, 0.5], [1, 0.5]])
    assert np.all(interpolated_point == [0, 0.5, 1])


def test_linear_nd():
    """
    In its simplest configuration this code should behave exactly the same as the scipy
    LinearNDInterpolator, so lets test that
    """

    # First set up 4 grid points and fill them randomly
    interpolation_points = {(0, 0): np.random.rand(2, 2),
                            (0, 1): np.random.rand(2, 2),
                            (1, 0): np.random.rand(2, 2),
                            (1, 1): np.random.rand(2, 2)}

    # Create UnstructuredInterpolator and LinearNDInterpolator with these points
    interpolator = UnstructuredInterpolator(interpolation_points)
    linear_nd = LinearNDInterpolator(list(interpolation_points.keys()),
                                     list(interpolation_points.values()))

    # Create some random coordinates in this space
    points = np.random.rand(10, 2)
    # And interpolate...
    interpolated_points = interpolator(points)
    linear_nd_points = linear_nd(points)

    # Check everything agrees to a reasonable precision
    assert np.all(np.abs(interpolated_points - linear_nd_points) < 1e-10)


def test_class_output():
    """
    The final test is to use the more useful functionality of interpolating between the
    outputs of a class member function. I will do this by interpolating between a
    number of numpy regulat grid interpolators and comparing to the output of the
    linear nd interpolator. Again this is a crazy use case, but is a good test.
    """

    x = np.linspace(0, 1, 11)
    # Create a bunch of random numbers to interpolate between
    rand_numbers = np.random.rand(4, 11, 11)

    # Create input for UnstructuredInterpolator
    interpolation_points = {(0, 0): RegularGridInterpolator((x, x), rand_numbers[0]),
                            (0, 1): RegularGridInterpolator((x, x), rand_numbers[1]),
                            (1, 0): RegularGridInterpolator((x, x), rand_numbers[2]),
                            (1, 1): RegularGridInterpolator((x, x), rand_numbers[3])}

    # Create some random points to evaluate our interpolators
    pts1 = np.random.rand(1, 2)
    pts2 = np.random.rand(10, 2)

    interpolator = UnstructuredInterpolator(interpolation_points)
    unsort_value = interpolator(pts1, pts2)

    interpolation_points = {(0, 0): rand_numbers[0],
                            (0, 1): rand_numbers[1],
                            (1, 0): rand_numbers[2],
                            (1, 1): rand_numbers[3]}

    # Perform the same operation by interpolating the values of the full numpy array
    linear_nd = LinearNDInterpolator(list(interpolation_points.keys()),
                                     list(interpolation_points.values()))
    array_out = linear_nd(pts1)
    # Then interpolate on this grid
    reg_interpolator = RegularGridInterpolator((x, x), array_out[0])
    lin_nd_val = reg_interpolator(pts2)

    # Check they give the same answer
    assert np.all(np.abs(unsort_value - lin_nd_val) < 1e-10)


if __name__ == '__main__':

    test_simple_interpolation()
    test_linear_nd()
    test_class_output()



"""collects converter functions from the various `geometry_converter_XXX` files in one
common module"""


from .geometry_converter_hex import (convert_geometry_hex1d_to_rect2d,
                                     convert_geometry_rect2d_back_to_hexe1d)

from .geometry_converter_astri import astri_to_2d_array, array_2d_to_astri
from .geometry_converter_chec import chec_to_2d_array, array_2d_to_chec



import numpy as np
from astropy import units as u
from astropy.time import Time
from ctapipe.instrument import TelescopeDescription
from ctapipe.io.eventsource import EventSource
from ctapipe.io.containers import TargetIODataContainer

__all__ = ['TargetIOEventSource']


class TargetIOEventSource(EventSource):
    """
    EventSource for the targetio unofficial data format, the data
    format used by cameras containing TARGET modules, such as CHEC for
    the GCT SST.

    Extract waveform information from `target_io` to store them
    into `ctapipe.io.containers`.

    This Extractor can fill either the R0 or R1 event container, depending on
    the file being read (The header of the file is checked for a flag
    indicating that it has had R1 calibration applied to it).

    This reader requires the TARGET libraries. The instructions to install
    these libraries can be found here:
    https://forge.in2p3.fr/projects/gct/wiki/Installing_CHEC_Software

    Attributes
    ----------
    _tio_reader : target_io.TargetIOEventReader()
        C++ event reader for TargetIO files. Handles the event building into
        an array of (n_pixels, n_samples) in C++, avoiding loops in Python
    _n_events : int
        number of events in the fits file
    _n_samples : int
        number of samples in the waveform
    _r0_samples : ndarray
        three dimensional array to store the R0 level waveform for each pixel
        (n_channels, n_pixels, n_samples)
    _r1_samples : ndarray
        three dimensional array to store the R1 level waveform for each pixel
        (n_channels, n_pixels, n_samples)
    _samples : ndarray
        pointer to the first index of either r0_samples or r1_samples
        (depending if the file has been R1 calibrated) for passing to
        TargetIO to be filled
    """

    def __init__(self, config=None, tool=None, **kwargs):
        super().__init__(config=config, tool=tool, **kwargs)
        try:
            import target_driver
            import target_io
            import target_calib
        except ImportError:
            msg = ("Cannot find TARGET libraries, please follow installation "
                   "instructions from https://forge.in2p3.fr/projects/gct/"
                   "wiki/Installing_CHEC_Software")
            self.log.error(msg)
            raise

        self._waveform_array_reader = target_io.WaveformArrayReader
        self._config_constr = target_calib.CameraConfiguration

        self._data = None
        self._event_index = None
        self._event_id = 0
        self._time_tack = None
        self._time_sec = None
        self._time_ns = None

        self._reader = self._waveform_array_reader(self.input_url, 2, 1)

        self._n_events = self._reader.fNEvents
        self._first_event_id = self._reader.fFirstEventID
        self._last_event_id = self._reader.fLastEventID
        self._obs_id = self._reader.fRunID
        n_modules = self._reader.fNModules
        n_pix = self._reader.fNPixels
        n_samples = self._reader.fNSamples
        self.camera_config = self._config_constr(self._reader.fCameraVersion)
        self._n_cells = self.camera_config.GetNCells()
        m = self.camera_config.GetMapping(n_modules == 1)

        self._optical_foclen = 2.283
        self._pixel_pos = np.vstack([m.GetXPixVector(), m.GetYPixVector()])
        self._refshape = np.zeros(10)  # TODO: Get correct values for CHEC-S
        self._refstep = 0  # TODO: Get correct values for CHEC-S
        self._time_slice = 0  # TODO: Get correct values for CHEC-S
        self._chec_tel = 0

        # Init fields
        self._r0_samples = None
        self._r1_samples = None
        self._first_cell_ids = np.zeros(n_pix, dtype=np.uint16)

        # Check if file is already r1 (Information obtained from a flag
        # in the file's header)
        is_r1 = self._reader.fR1
        if is_r1:
            self._r1_samples = np.zeros(
                (1, n_pix, n_samples),
                dtype=np.float32
            )
            self._get_tio_event = self._reader.GetR1Event
            self._samples = self._r1_samples[0]
        else:
            self._r0_samples = np.zeros(
                (1, n_pix, n_samples),
                dtype=np.uint16
            )
            self._get_tio_event = self._reader.GetR0Event
            self._samples = self._r0_samples[0]

        self._init_container()

    @staticmethod
    def is_compatible(file_path):
        return file_path.endswith('.tio')

    def _init_container(self):
        """
        Prepare the ctapipe event container, and fill it with the information
        that does not change with event, including the instrument information.
        """
        chec_tel = 0

        data = TargetIODataContainer()
        data.meta['origin'] = "targetio"

        data.meta['input'] = self.input_url
        data.meta['max_events'] = self.max_events

        # Instrument information
        pix_pos = self._pixel_pos * u.m
        foclen = self._optical_foclen * u.m
        teldesc = TelescopeDescription.guess(*pix_pos, foclen)
        data.inst.subarray.tels[chec_tel] = teldesc

        self._data = data

    def _update_container(self):
        """
        Update the ctapipe event containers with the information from the
        current event being pointed to in TargetIO.
        """
        data = self._data
        chec_tel = 0

        obs_id = self._obs_id
        event_id = self._event_id
        tels = {self._chec_tel}

        data.r0.obs_id = obs_id
        data.r0.event_id = event_id
        data.r0.tels_with_data = tels
        data.r1.obs_id = obs_id
        data.r1.event_id = event_id
        data.r1.tels_with_data = tels
        data.dl0.obs_id = obs_id
        data.dl0.event_id = event_id
        data.dl0.tels_with_data = tels

        data.trig.tels_with_trigger = [chec_tel]

        data.meta['tack'] = self._time_tack
        data.meta['sec'] = self._time_sec
        data.meta['ns'] = self._time_ns
        data.trig.gps_time = Time(self._time_sec * u.s, self._time_ns * u.ns,
                                  format='unix', scale='utc', precision=9)

        data.count = self._event_index

        data.r0.tel.clear()
        data.r1.tel.clear()
        data.dl0.tel.clear()
        data.dl1.tel.clear()
        data.mc.tel.clear()
        data.targetio.tel.clear()

        # load the data per telescope/chan
        data.r0.tel[chec_tel].waveform = self._r0_samples
        data.r1.tel[chec_tel].waveform = self._r1_samples

        # Load the TargetIO specific data per telescope/chan
        data.targetio.tel[chec_tel].first_cell_ids = self._first_cell_ids
        data.r0.tel[chec_tel].num_samples = self._samples.shape[-1]

        # Some information that currently exists in the mc container, but is
        # useful for real data (essentially the reference pulse shape,
        # which may be used in charge extraction methods)
        data.mc.tel[chec_tel].reference_pulse_shape = self._refshape
        data.mc.tel[chec_tel].meta['refstep'] = self._refstep
        data.mc.tel[chec_tel].time_slice = self._time_slice

    @property
    def _current_event_index(self):
        return self._event_index

    @_current_event_index.setter
    def _current_event_index(self, val):
        """
        Setting the event index will cause the event to be saught from
        TargetIO, and the Containers to point to
        the correct event. The ctapipe event containers are then updated
        with this new event's information.
        """
        self._event_index = val
        self._get_tio_event(val, self._samples, self._first_cell_ids)
        self._event_id = self._reader.fCurrentEventID
        self._time_tack = self._reader.fCurrentTimeTack
        self._time_sec = self._reader.fCurrentTimeSec
        self._time_ns = self._reader.fCurrentTimeNs
        self._update_container()

    def _generator(self):
        for self._current_event_index in range(self._n_events):
            yield self._data
        return

    def __len__(self):
        num = self._n_events
        if self.max_events and self.max_events < num:
            num = self.max_events
        return num

    def _get_event_by_index(self, index):
        self._current_event_index = index
        return self._data

    def _get_event_by_id(self, event_id):
        if ((event_id < self._first_event_id) |
                (event_id > self._last_event_id)):
            raise IndexError("Event id {} not found in file".format(event_id))
        index = self._reader.GetEventIndex(event_id)
        return self._get_event_by_index(index)



"""
Functions to retrieve and interpolate atmosphere profiles.
"""
import numpy as np
from astropy.units import Quantity
from scipy.interpolate import interp1d

from ctapipe.utils import get_table_dataset

__all__ = ['get_atmosphere_profile_table', 'get_atmosphere_profile_functions']


def get_atmosphere_profile_table(atmosphere_name='paranal'):
    """
    Get an atmosphere profile table

    Parameters
    ----------
    atmosphere_name: str
        identifier of atmosphere profile

    Returns
    -------
    astropy.table.Table  containing atmosphere profile with at least columns
    'altitude' (m), and 'thickness' (g cm-2) as well as others.

    """
    table_name = '{}.atmprof'.format(atmosphere_name)
    table = get_table_dataset(table_name=table_name,
                              role='dl0.arr.svc.atmosphere')
    return table


def get_atmosphere_profile_functions(atmosphere_name="paranal",
                                     with_units=True):
    """
    Gives atmospheric profile as a continuous function thickness(
    altitude), and it's inverse altitude(thickness)  in m and g/cm^2
    respectively.

    Parameters
    ----------
    atmosphere_name: str
        identifier of atmosphere profile
    with_units: bool
       if true, return functions that accept and return unit quantities.
       Otherwise assume units are 'm' and 'g cm-2'

    Returns
    -------
    functions: thickness(alt), alt(thickness)
    """
    tab = get_atmosphere_profile_table(atmosphere_name)
    alt = tab['altitude'].to('m')
    thick = (tab['thickness']).to("g cm-2")

    alt_to_thickness = interp1d(x=np.array(alt), y=np.array(thick))
    thickness_to_alt = interp1d(x=np.array(thick), y=np.array(alt))

    if with_units:
        def thickness(a):
            return Quantity(alt_to_thickness(a.to('m')), 'g cm-2')

        def altitude(a):
            return Quantity(thickness_to_alt(a.to('g cm-2')), 'm')

        return thickness, altitude

    return alt_to_thickness, thickness_to_alt



from math import log10, sqrt
from os import makedirs
from os.path import dirname, exists

import numpy as np
from scipy.stats import binned_statistic as bs
from tables import open_file

from ctapipe.core import Component
from ctapipe.core.traits import Int, Bool

__all__ = ['ChargeResolutionCalculator']


class ChargeResolutionCalculator(Component):
    """
    Class to handle the calculation of Charge Resolution.

    Attributes
    ----------
    max_pe : int
        Maximum pe to calculate the charge resolution up to.
    sum_dict : dict
        Dictionary to store the running sum for each true charge.
    n_dict : dict
        Dictionary to store the running number for each true charge.
    variation_hist_nbins : float
        Number of bins for the variation histogram.
    variation_hist_range : list
        X and Y range for the variation histogram.
    variation_hist : `np.histogram2d`
    variation_xedges : ndarray
        Edges of the X bins for the variation histogram.
    variation_yedges : ndarray
        Edges of the Y bins for the variation histogram.
    """
    max_pe = Int(2000, help='Maximum pe to calculate the charge resolution '
                            'up to').tag(config=True)
    binning = Int(60, allow_none=True,
                  help='Number of bins for the Charge Resolution. If None, '
                       'no binning is performed.').tag(config=True)
    log_bins = Bool(True, help='Bin the x axis linearly instead of '
                               'logarithmic.').tag(config=True)

    def __init__(self, config=None, tool=None, **kwargs):
        """
        Calculator of charge resolution.

        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        kwargs
        """
        super().__init__(config=config, parent=tool, **kwargs)

        self.sum_array = np.zeros(self.max_pe)
        self.n_array = np.zeros(self.max_pe)
        self.sum_dict = {}
        self.n_dict = {}

        self.variation_hist_nbins = log10(self.max_pe) * 50
        self.variation_hist_range = [[log10(1), log10(self.max_pe)],
                                     [log10(1), log10(self.max_pe)]]
        h, xedges, yedges = np.histogram2d([np.nan], [np.nan],
                                           bins=self.variation_hist_nbins,
                                           range=self.variation_hist_range)
        self.variation_hist = h
        self.variation_xedges = xedges
        self.variation_yedges = yedges

        self.storage_arrays = ['max_pe', 'sum_array', 'n_array',
                               'variation_hist_nbins', 'variation_hist_range',
                               'variation_hist', 'variation_xedges',
                               'variation_yedges']

    def add_charges(self, true_charge, measured_charge):
        """
        Fill the class parameters with a numpy array of true charge and
        measured (calibrated) charge from an event. The two arrays must be the
        same size.

        Parameters
        ----------
        true_charge : ndarray
            Array of true (MC) charge.
            Obtained from event.mc.tel[telid].image[channel]
        measured_charge : ndarray
            Array of measured (dl1 calibrated) charge.
            Obtained from event.mc.tel[tel_id].photo_electron_image
        """
        above_0 = (measured_charge > 0) & (true_charge > 0)
        x = true_charge[above_0]
        y = measured_charge[above_0]
        h, _, _ = np.histogram2d(np.log10(y),
                                 np.log10(x),
                                 bins=self.variation_hist_nbins,
                                 range=self.variation_hist_range)
        self.variation_hist += h

        in_range = (true_charge > 0) & (true_charge <= self.max_pe)
        true_q = true_charge[in_range]
        measured_q = measured_charge[in_range]
        np.add.at(self.sum_array, true_q - 1, np.power(measured_q - true_q, 2))
        np.add.at(self.n_array, true_q - 1, 1)

    def get_charge_resolution(self):
        """
        Calculate and obtain the charge resolution graph arrays.

        Returns
        -------
        true_charge : ndarray
            The X axis true charges.
        chargeres : ndarray
            The Y axis charge resolution values.
        chargeres_error : ndarray
            The error on the charge resolution.
        scaled_chargeres : ndarray
            The Y axis charge resolution divided by the Goal.
        scaled_chargeres_error : ndarray
            The error on the charge resolution divided by the Goal.
        """
        self.log.debug('[chargeres] Calculating charge resolution')

        n_1 = self.n_array > 0
        n = self.n_array[n_1]
        true = (np.arange(self.max_pe) + 1)[n_1]
        sum_ = self.sum_array[n_1]

        res = np.sqrt((sum_ / n) + true) / true
        res_error = res * (1 / np.sqrt(2 * n))

        scale = self.goal(true)
        scaled_res = res / scale
        scaled_res_error = res_error / scale

        if self.binning is not None:
            x = true
            if self.log_bins:
                x = np.log10(true)

            def binning(array):
                return bs(x, array, 'mean', bins=self.binning)

            def sum_errors(array):
                return np.sqrt(np.sum(np.power(array, 2))) / array.size

            def bin_errors(array):
                return bs(x, array, sum_errors, bins=self.binning)

            true, _, _ = binning(true)
            res, _, _ = binning(res)
            res_error, _, _ = bin_errors(res_error)
            scaled_res, _, _ = binning(scaled_res)
            scaled_res_error, _, _ = bin_errors(scaled_res_error)

        return true, res, res_error, scaled_res, scaled_res_error

    @staticmethod
    def limit_curves(npe, n_nsb, n_add, enf, sigma2):
        """
        Equation for calculating the Goal and Requirement curves, as defined
        in SCI-MC/121113.
        https://portal.cta-observatory.org/recordscentre/Records/SCI/
        SCI-MC/measurment_errors_system_performance_1YQCBC.pdf

        Parameters
        ----------
        npe : ndarray
            Number of photoeletrons (variable).
        n_nsb : float
            Number of NSB photons.
        n_add : float
            Number of photoelectrons from additional noise sources.
        enf : float
            Excess noise factor.
        sigma2 : float
            Percentage ofmultiplicative errors.
        """
        return (np.sqrt((n_nsb + n_add) + np.power(enf, 2) * npe +
                        np.power(sigma2 * npe, 2)) / npe).astype(float)

    @staticmethod
    def requirement(npe):
        """
        CTA requirement curve.

        Parameters
        ----------
        npe : ndarray
            Number of photoeletrons (variable).
        """
        n_nsb = sqrt(4.0 + 3.0)
        n_add = 0
        enf = 1.2
        sigma2 = 0.1
        defined_npe = 1000

        # If npe is not an array, temporarily convert it to one
        npe = np.array([npe])

        lc = ChargeResolutionCalculator.limit_curves
        requirement = lc(npe, n_nsb, n_add, enf, sigma2)
        requirement[npe > defined_npe] = np.nan

        return requirement[0]

    @staticmethod
    def goal(npe):
        """
        CTA goal curve.

        Parameters
        ----------
        npe : ndarray
            Number of photoeletrons (variable).
        """
        n_nsb = 2
        n_add = 0
        enf = 1.1152
        sigma2 = 0.05
        defined_npe = 2000

        # If npe is not an array, temporarily convert it to one
        npe = np.array([npe])

        lc = ChargeResolutionCalculator.limit_curves
        goal = lc(npe, n_nsb, n_add, enf, sigma2)
        goal[npe > defined_npe] = np.nan

        return goal[0]

    @staticmethod
    def poisson(npe):
        """
        Poisson limit curve.

        Parameters
        ----------
        npe : ndarray
            Number of photoeletrons (variable).
        """
        # If npe is not an array, temporarily convert it to one
        npe = np.array([npe])
        poisson = np.sqrt(npe) / npe

        return poisson[0]

    def save(self, path):
        output_dir = dirname(path)
        if not exists(output_dir):
            self.log.info("[output] Creating directory: {}".format(output_dir))
            makedirs(output_dir)
        self.log.info("Saving Charge Resolution file: {}".format(path))

        with open_file(path, mode="w", title="ChargeResolutionFile") as f:
            group = f.create_group("/", 'ChargeResolution', '')
            for arr in self.storage_arrays:
                f.create_array(group, arr, getattr(self, arr), arr)

    def load(self, path):
        self.log.info("Loading Charge Resolution file: {}".format(path))
        with open_file(path, mode="r") as f:
            for arr in self.storage_arrays:
                setattr(self, arr, f.get_node("/ChargeResolution", arr).read())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Backward compatibility function for reading hessio files.

"""
import logging

from .hessioeventsource import HESSIOEventSource
from astropy.utils.decorators import deprecated

logger = logging.getLogger(__name__)


__all__ = [
    'hessio_event_source',
]


@deprecated(0.5, message="prefer the use of an EventSource or "
                         "EventSourceFactory")
def hessio_event_source(url, **kwargs):
    """
    emulate the old `hessio_event_source` generator, using the new
    `HESSIOEventSource` class.  It is preferred to use `HESSIOEventSource` or
    `event_source`, this is only for backward compatibility.


    Parameters
    ----------
    url : str
        path to file to open
    kwargs:
        extra parameters to pass to HESSIOEventSource



    Returns
    -------
    generator:
        a `HESSIOEventSource` wrapped in a generator (for backward
        compatibility)

    """

    reader = HESSIOEventSource(input_url=url, **kwargs)

    return (x for x in reader)




"""
Charge extraction algorithms to reduce the image to one value per pixel
"""

__all__ = ['ChargeExtractorFactory', 'FullIntegrator', 'SimpleIntegrator',
           'GlobalPeakIntegrator', 'LocalPeakIntegrator',
           'NeighbourPeakIntegrator', 'AverageWfPeakIntegrator']


from abc import abstractmethod
import numpy as np
from traitlets import Int, CaselessStrEnum, Float
from ctapipe.core import Component, Factory
from ctapipe.utils.neighbour_sum import get_sum_array


class ChargeExtractor(Component):

    def __init__(self, config=None, tool=None, **kwargs):
        """
        Base component to handle the extraction of charge from an image cube.

        Attributes
        ----------
        neighbours : ndarray
            2D array where each row is [pixel index, one neighbour
            of that pixel].
            Changes per telescope.
            Can be obtained from
            `ctapipe.instrument.CameraGeometry.neighbor_matrix_where`.

        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool or None
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        kwargs
        """
        super().__init__(config=config, parent=tool, **kwargs)

        self.neighbours = None

    @staticmethod
    def requires_neighbours():
        """
        Method used for callers of the ChargeExtractor to know if the extractor
        requires knowledge of the pixel neighbours

        Returns
        -------
        bool
        """
        return False

    def check_neighbour_set(self):
        """
        Check if the pixel neighbours has been set for the extractor

        Raises
        -------
        ValueError
            If neighbours has not been set
        """
        if self.requires_neighbours():
            if self.neighbours is None:
                self.log.exception("neighbours attribute must be set")
                raise ValueError()

    @abstractmethod
    def get_peakpos(self, waveforms):
        """
        Get the peak position from the waveforms

        Parameters
        ----------
        waveforms : ndarray
            Waveforms stored in a numpy array of shape
            (n_chan, n_pix, n_samples).

        Returns
        -------
        peakpos : ndarray
            Numpy array of the peak position for each pixel. 
            Has shape of (n_chan, n_pix).

        """

    @abstractmethod
    def extract_charge(self, waveforms):
        """
        Call the relevant functions to fully extract the charge for the
        particular extractor.

        Parameters
        ----------
        waveforms : ndarray
            Waveforms stored in a numpy array of shape
            (n_chan, n_pix, n_samples).

        Returns
        -------
        charge : ndarray
            Extracted charge stored in a numpy array of shape (n_chan, n_pix).
        window : ndarray
            Bool numpy array defining the samples included in the integration
            window.
        """


class Integrator(ChargeExtractor):

    def __init__(self, config=None, tool=None, **kwargs):
        """
        Base component for charge extractors that perform integration.

        Attributes
        ----------
        neighbours : list
            List of neighbours for each pixel. Changes per telescope.

        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        kwargs
        """
        super().__init__(config=config, tool=tool, **kwargs)

    @staticmethod
    def check_window_width_and_start(n_samples, start, width):
        """
        Check that the combination of window width and start positions fit
        within the readout window.

        Parameters
        ----------
        n_samples : int
            Number of samples in the waveform
        start : ndarray
            Numpy array containing the window start for each pixel. Shape =
            (n_chan, n_pix)
        width : ndarray
            Numpy array containing the window width for each pixel. Shape =
            (n_chan, n_pix)

        """
        width[width > n_samples] = n_samples
        start[start < 0] = 0
        sum_check = start + width > n_samples
        start[sum_check] = n_samples - width[sum_check]

    @abstractmethod
    def get_peakpos(self, waveforms):
        """"""

    @abstractmethod
    def _get_window_start(self, waveforms, peakpos):
        """
        Get the starting point for the integration window

        Parameters
        ----------
        waveforms : ndarray
            Waveforms stored in a numpy array of shape
            (n_chan, n_pix, n_samples).
        peakpos : ndarray
            Numpy array containing the peak position for each pixel. 
            Shape = (n_chan, n_pix)

        Returns
        -------
        start : ndarray
            Numpy array containing the window start for each pixel. 
            Shape = (n_chan, n_pix)

        """

    @abstractmethod
    def _get_window_width(self, waveforms):
        """
        Get the width of the integration window

        Parameters
        ----------
        waveforms : ndarray
            Waveforms stored in a numpy array of shape
            (n_chan, n_pix, n_samples).

        Returns
        -------
        width : ndarray
            Numpy array containing the window width for each pixel. 
            Shape = (n_chan, n_pix)

        """

    def get_start_and_width(self, waveforms, peakpos):
        """
        Obtain the numpy arrays containing the start and width for the
        integration window for each pixel.

        Parameters
        ----------
        waveforms : ndarray
            Waveforms stored in a numpy array of shape
            (n_chan, n_pix, n_samples).
        peakpos : ndarray
            Numpy array of the peak position for each pixel. 
            Has shape of (n_chan, n_pix).

        Returns
        -------
        w_start : ndarray
            Numpy array containing the Start sample of integration window.
            Shape: (n_chan, n_pix).
        w_width : ndarray
            Numpy array containing the window size of integration window.
            Shape (n_chan, n_pix).
        """
        w_start = self._get_window_start(waveforms, peakpos)
        w_width = self._get_window_width(waveforms)
        n_samples = waveforms.shape[2]
        self.check_window_width_and_start(n_samples, w_start, w_width)
        return w_start, w_width

    @staticmethod
    def get_window(waveforms, start, width):
        """
        Build the a numpy array of bools defining the integration window.

        Parameters
        ----------
        waveforms : ndarray
            Waveforms stored in a numpy array of shape
            (n_chan, n_pix, n_samples).
        start : ndarray
            Numpy array containing the Start sample of integration window.
            Shape: (n_chan, n_pix).
        width : ndarray
            Numpy array containing the window size of integration window.
            Shape (n_chan, n_pix).

        Returns
        -------
        integration_window : ndarray
            Numpy array containing True where the samples lay within the
            integration window, and False where the samples lay outside. Has
            shape of (n_chan, n_pix, n_samples).

        """
        end = start + width

        # Obtain integration window using the indices of the waveforms array
        ind = np.indices(waveforms.shape)[2]
        integration_window = (ind >= start[..., None]) & (ind < end[..., None])
        return integration_window

    @staticmethod
    def extract_from_window(waveforms, window):
        """
        Extract the charge but applying an intregration window to the 
        waveforms.

        Parameters
        ----------
        waveforms : ndarray
            Waveforms stored in a numpy array of shape
            (n_chan, n_pix, n_samples).
        window : ndarray
            Numpy array containing True where the samples lay within the
            integration window, and False where the samples lay outside. Has
            shape of (n_chan, n_pix, n_samples).

        Returns
        -------
        charge : ndarray
            Extracted charge stored in a numpy array of shape (n_chan, n_pix).
        """
        windowed = np.ma.array(waveforms, mask=~window)
        charge = windowed.sum(2).data
        return charge

    def get_window_from_waveforms(self, waveforms):
        """
        Consolidating function to obtain the window and peakpos given 
        a waveform.

        Parameters
        ----------
        waveforms : ndarray
            Waveforms stored in a numpy array of shape
            (n_chan, n_pix, n_samples).

        Returns
        -------
        window : ndarray
            Numpy array containing True where the samples lay within the
            integration window, and False where the samples lay outside. Has
            shape of (n_chan, n_pix, n_samples).
        peakpos : ndarray
            Numpy array of the peak position for each pixel. 
            Has shape of (n_chan, n_pix).

        """
        peakpos = self.get_peakpos(waveforms)
        start, width = self.get_start_and_width(waveforms, peakpos)
        window = self.get_window(waveforms, start, width)
        return window, peakpos

    def extract_charge(self, waveforms):
        window, peakpos = self.get_window_from_waveforms(waveforms)
        charge = self.extract_from_window(waveforms, window)
        return charge, peakpos, window


class FullIntegrator(Integrator):

    def __init__(self, config=None, tool=None, **kwargs):
        """
        Charge extractor that integrates the entire waveform.

        Attributes
        ----------
        neighbours : list
            List of neighbours for each pixel. Changes per telescope.

        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        kwargs
        """
        super().__init__(config=config, tool=tool, **kwargs)

    def _get_window_start(self, waveforms, peakpos):
        nchan, npix, nsamples = waveforms.shape
        return np.zeros((nchan, npix), dtype=np.intp)

    def _get_window_width(self, waveforms):
        nchan, npix, nsamples = waveforms.shape
        return np.full((nchan, npix), nsamples, dtype=np.intp)

    def get_peakpos(self, waveforms):
        nchan, npix, nsamples = waveforms.shape
        return np.zeros((nchan, npix), dtype=np.intp)


class WindowIntegrator(Integrator):
    window_shift = Int(3, help='Define the shift of the integration window '
                               'from the peakpos '
                               '(peakpos - shift)').tag(config=True)
    window_width = Int(7, help='Define the width of the integration '
                               'window').tag(config=True)

    def __init__(self, config=None, tool=None, **kwargs):
        """
        Base component for charge extractors that perform integration within
        a window.

        Attributes
        ----------
        neighbours : list
            List of neighbours for each pixel. Changes per telescope.

        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        kwargs
        """
        super().__init__(config=config, tool=tool, **kwargs)

    def get_peakpos(self, waveforms):
        return self._obtain_peak_position(waveforms)

    def _get_window_start(self, waveforms, peakpos):
        return peakpos - self.window_shift

    def _get_window_width(self, waveforms):
        nchan, npix, nsamples = waveforms.shape
        return np.full((nchan, npix), self.window_width, dtype=np.intp)

    @abstractmethod
    def _obtain_peak_position(self, waveforms):
        """
        Find the peak to define integration window around.

        Parameters
        ----------
        waveforms : ndarray
            Waveforms stored in a numpy array of shape
            (n_chan, n_pix, n_samples).

        Returns
        -------
        peakpos : ndarray
            Numpy array of the peak position for each pixel. Has shape of
            (n_chan, n_pix).

        """


class SimpleIntegrator(WindowIntegrator):
    t0 = Int(0, help='Define the peak position for all '
                     'pixels').tag(config=True)

    def __init__(self, config=None, tool=None, **kwargs):
        """
        Charge extractor that integrates within a window defined by the user.

        Attributes
        ----------
        neighbours : list
            List of neighbours for each pixel. Changes per telescope.

        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        kwargs
        """
        super().__init__(config=config, tool=tool, **kwargs)

    def _obtain_peak_position(self, waveforms):
        nchan, npix, nsamples = waveforms.shape
        return np.full((nchan, npix), self.t0, dtype=np.intp)


class PeakFindingIntegrator(WindowIntegrator):
    sig_amp_cut_HG = Float(None, allow_none=True,
                           help='Define the cut above which a sample is '
                                'considered as significant for PeakFinding '
                                'in the HG channel').tag(config=True)
    sig_amp_cut_LG = Float(None, allow_none=True,
                           help='Define the cut above which a sample is '
                                'considered as significant for PeakFinding '
                                'in the LG channel').tag(config=True)

    def __init__(self, config=None, tool=None, **kwargs):
        """
        Base component for charge extractors that perform integration within
        a window defined around a peak position.

        Attributes
        ----------
        neighbours : list
            List of neighbours for each pixel. Changes per telescope.

        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        kwargs
        """
        super().__init__(config=config, tool=tool, **kwargs)
        self._sig_channel = None
        self._sig_pixels = None

    # Extract significant entries
    def _extract_significant_entries(self, waveforms):
        """
        Obtain the samples that the user has specified as significant.

        Parameters
        ----------
        waveforms : ndarray
            Waveforms stored in a numpy array of shape
            (n_chan, n_pix, n_samples).

        Returns
        -------
        significant_samples : masked ndarray
            Identical to waveforms, except the insignificant samples are
            masked.

        """
        nchan, npix, nsamples = waveforms.shape
        if self.sig_amp_cut_LG or self.sig_amp_cut_HG:
            sig_entries = np.ones(waveforms.shape, dtype=bool)
            if self.sig_amp_cut_HG:
                sig_entries[0] = waveforms[0] > self.sig_amp_cut_HG
            if nchan > 1 and self.sig_amp_cut_LG:
                sig_entries[1] = waveforms[1] > self.sig_amp_cut_LG
            self._sig_pixels = np.any(sig_entries, axis=2)
            self._sig_channel = np.any(self._sig_pixels, axis=1)
            if not self._sig_channel[0]:
                self.log.error("sigamp excludes all values in HG channel")
            return np.ma.array(waveforms, mask=~sig_entries)
        else:
            self._sig_channel = np.ones(nchan, dtype=bool)
            self._sig_pixels = np.ones((nchan, npix), dtype=bool)
            return waveforms

    @abstractmethod
    def _obtain_peak_position(self, waveforms):
        """"""


class GlobalPeakIntegrator(PeakFindingIntegrator):
    """
    Charge extractor that defines an integration window about the global
    peak in the image.

    Attributes
    ----------
    neighbours : list
        List of neighbours for each pixel. Changes per telescope.

    Parameters
    ----------
    config : traitlets.loader.Config
        Configuration specified by config file or cmdline arguments.
        Used to set traitlet values.
        Set to None if no configuration to pass.
    tool : ctapipe.core.Tool
        Tool executable that is calling this component.
        Passes the correct logger to the component.
        Set to None if no Tool to pass.
    kwargs
    """

    def __init__(self, config=None, tool=None, **kwargs):

        super().__init__(config=config, tool=tool, **kwargs)

    def _obtain_peak_position(self, waveforms):
        nchan, npix, nsamples = waveforms.shape
        significant_samples = self._extract_significant_entries(waveforms)
        max_t = significant_samples.argmax(2)
        max_s = significant_samples.max(2)

        peakpos = np.zeros((nchan, npix), dtype=np.int)
        peakpos[0, :] = np.round(np.average(max_t[0], weights=max_s[0]))
        if nchan > 1:
            if self._sig_channel[1]:
                peakpos[1, :] = np.round(
                    np.average(max_t[1], weights=max_s[1]))
            else:
                self.log.info("LG not significant, using HG for peak finding "
                              "instead")
                peakpos[1, :] = peakpos[0]
        return peakpos


class LocalPeakIntegrator(PeakFindingIntegrator):
    """
     Charge extractor that defines an integration window about the local
     peak in each pixel.

     Attributes
     ----------
     neighbours : list
         List of neighbours for each pixel. Changes per telescope.

     Parameters
     ----------
     config : traitlets.loader.Config
         Configuration specified by config file or cmdline arguments.
         Used to set traitlet values.
         Set to None if no configuration to pass.
     tool : ctapipe.core.Tool
         Tool executable that is calling this component.
         Passes the correct logger to the component.
         Set to None if no Tool to pass.
     kwargs
     """

    def __init__(self, config=None, tool=None, **kwargs):
        super().__init__(config=config, tool=tool, **kwargs)

    def _obtain_peak_position(self, waveforms):
        nchan, npix, nsamples = waveforms.shape
        significant_samples = self._extract_significant_entries(waveforms)
        peakpos = np.full((nchan, npix), significant_samples.argmax(2),
                          dtype=np.int)
        sig_pix = self._sig_pixels
        if nchan > 1:  # If the LG is not significant, use the HG peakpos
            peakpos[1] = np.where(sig_pix[1] < sig_pix[0],
                                  peakpos[0], peakpos[1])
        return peakpos


class NeighbourPeakIntegrator(PeakFindingIntegrator):
    """
    Charge extractor that defines an integration window defined by the 
    peaks in the neighbouring pixels.

    Attributes
    ----------
    neighbours : list
        List of neighbours for each pixel. Changes per telescope.

    Parameters
    ----------
    config : traitlets.loader.Config
        Configuration specified by config file or cmdline arguments.
        Used to set traitlet values.
        Set to None if no configuration to pass.
    tool : ctapipe.core.Tool
        Tool executable that is calling this component.
        Passes the correct logger to the component.
        Set to None if no Tool to pass.
    kwargs
    """
    lwt = Int(0, help='Weight of the local pixel (0: peak from neighbours '
                      'only, 1: local pixel counts as much '
                      'as any neighbour').tag(config=True)

    def __init__(self, config=None, tool=None, **kwargs):
        super().__init__(config=config, tool=tool, **kwargs)

    @staticmethod
    def requires_neighbours():
        return True

    def _obtain_peak_position(self, waveforms):
        shape = waveforms.shape
        significant_samples = self._extract_significant_entries(waveforms)
        sig_sam = significant_samples.astype(np.float32)
        sum_data = np.zeros_like(sig_sam)
        n = self.neighbours.astype(np.uint16)
        get_sum_array(sig_sam, sum_data, *shape, n, n.shape[0], self.lwt)
        return sum_data.argmax(2).astype(np.int)


class AverageWfPeakIntegrator(PeakFindingIntegrator):
    """
    Charge extractor that defines an integration window defined by the 
    average waveform across all pixels.

    Attributes
    ----------
    neighbours : list
        List of neighbours for each pixel. Changes per telescope.

    Parameters
    ----------
    config : traitlets.loader.Config
        Configuration specified by config file or cmdline arguments.
        Used to set traitlet values.
        Set to None if no configuration to pass.
    tool : ctapipe.core.Tool
        Tool executable that is calling this component.
        Passes the correct logger to the component.
        Set to None if no Tool to pass.
    kwargs
    """

    def __init__(self, config=None, tool=None, **kwargs):
        super().__init__(config=config, tool=tool, **kwargs)

    def _obtain_peak_position(self, waveforms):
        nchan, npix, nsamples = waveforms.shape
        significant_samples = self._extract_significant_entries(waveforms)
        peakpos = np.zeros((nchan, npix), dtype=np.int)
        avg_wf = np.mean(significant_samples, axis=1)
        peakpos += np.argmax(avg_wf, axis=1)[:, None]
        return peakpos


class ChargeExtractorFactory(Factory):
    """
    Factory to obtain a ChargeExtractor.
    """
    base = ChargeExtractor
    default = 'NeighbourPeakIntegrator'
    custom_product_help = 'Charge extraction scheme to use.'



import logging

import numpy as np
from astropy import log
from astropy import units as u
from astropy.utils.decorators import deprecated

from ctapipe.coordinates import CameraFrame, NominalFrame, HorizonFrame
from ctapipe.image.cleaning import tailcuts_clean
from ctapipe.image.muon.features import ring_containment
from ctapipe.image.muon.features import ring_completeness
from ctapipe.image.muon.features import npix_above_threshold
from ctapipe.image.muon.features import npix_composing_ring
from ctapipe.image.muon.muon_integrator import MuonLineIntegrate
from ctapipe.image.muon.muon_ring_finder import ChaudhuriKunduRingFitter

logger = logging.getLogger(__name__)


def analyze_muon_event(event):
    """
    Generic muon event analyzer. 

    Parameters
    ----------
    event : ctapipe dl1 event container


    Returns
    -------
    muonringparam, muonintensityparam : MuonRingParameter 
    and MuonIntensityParameter container event

    """

    names = ['LST:LSTCam', 'MST:NectarCam', 'MST:FlashCam', 'MST-SCT:SCTCam',
             'SST-1M:DigiCam', 'SST-GCT:CHEC', 'SST-ASTRI:ASTRICam', 'SST-ASTRI:CHEC']
    tail_cuts = [(5, 7), (5, 7), (10, 12), (5, 7),
                (5, 7), (5, 7), (5, 7), (5, 7)]  # 10, 12?
    impact = [(0.2, 0.9), (0.1, 0.95), (0.2, 0.9), (0.2, 0.9),
              (0.1, 0.95), (0.1, 0.95), (0.1, 0.95), (0.1, 0.95)] * u.m
    ringwidth = [(0.04, 0.08), (0.02, 0.1), (0.01, 0.1), (0.02, 0.1),
                 (0.01, 0.5), (0.02, 0.2), (0.02, 0.2), (0.02, 0.2)] * u.deg
    total_pix = [1855., 1855., 1764., 11328., 1296., 2048., 2368., 2048]
    # 8% (or 6%) as limit
    min_pix = [148., 148., 141., 680., 104., 164., 142., 164]
    # Need to either convert from the pixel area in m^2 or check the camera specs
    ang_pixel_width = [0.1, 0.2, 0.18, 0.067, 0.24, 0.2, 0.17, 0.2, 0.163] * u.deg
    # Found from TDRs (or the pixel area)
    hole_rad = [0.308 * u.m, 0.244 * u.m, 0.244 * u.m,
                4.3866 * u.m, 0.160 * u.m, 0.130 * u.m,
                0.171 * u.m, 0.171 * u.m]  # Assuming approximately spherical hole
    cam_rad = [2.26, 3.96, 3.87, 4., 4.45, 2.86, 5.25, 2.86] * u.deg
    # Above found from the field of view calculation
    sec_rad = [0. * u.m, 0. * u.m, 0. * u.m, 2.7 * u.m,
               0. * u.m, 1. * u.m, 1.8 * u.m, 1.8 * u.m]
    sct = [False, False, False, True, False, True, True, True]
    # Added cleaning here. All these options should go to an input card
    cleaning = True

    muon_cuts = {'Name': names, 'tail_cuts': tail_cuts, 'Impact': impact,
                 'RingWidth': ringwidth, 'total_pix': total_pix,
                 'min_pix': min_pix, 'CamRad': cam_rad, 'SecRad': sec_rad,
                 'SCT': sct, 'AngPixW': ang_pixel_width, 'HoleRad': hole_rad}
    logger.debug(muon_cuts)

    muonringlist = []  # [None] * len(event.dl0.tels_with_data)
    muonintensitylist = []  # [None] * len(event.dl0.tels_with_data)
    tellist = []
    muon_event_param = {'TelIds': tellist,
                        'MuonRingParams': muonringlist,
                        'MuonIntensityParams': muonintensitylist}

    for telid in event.dl0.tels_with_data:

        logger.debug("Analysing muon event for tel %d", telid)
        image = event.dl1.tel[telid].image[0]

        # Get geometry
        teldes = event.inst.subarray.tel[telid]
        geom = teldes.camera
        x, y = geom.pix_x, geom.pix_y

        dict_index = muon_cuts['Name'].index(str(teldes))
        logger.debug('found an index of %d for camera %d',
                     dict_index, geom.cam_id)

        tailcuts = muon_cuts['tail_cuts'][dict_index]
        logger.debug("Tailcuts are %s", tailcuts)

        clean_mask = tailcuts_clean(geom, image, picture_thresh=tailcuts[0],
                                    boundary_thresh=tailcuts[1])
        camera_coord = CameraFrame(
            x=x, y=y,
            focal_length=teldes.optics.equivalent_focal_length,
            rotation=geom.pix_rotation
        )

        # TODO: correct this hack for values over 90
        altval = event.mcheader.run_array_direction[1]
        if altval > np.pi / 2.:
            altval = np.pi / 2.

        altaz = HorizonFrame(alt=altval * u.rad,
                             az=event.mcheader.run_array_direction[0] * u.rad)
        nom_coord = camera_coord.transform_to(
            NominalFrame(array_direction=altaz, pointing_direction=altaz)
        )
        x = nom_coord.x.to(u.deg)
        y = nom_coord.y.to(u.deg)

        if(cleaning):
            img = image * clean_mask
        else:
            img = image

        muonring = ChaudhuriKunduRingFitter(None)

        logger.debug("img: %s mask: %s, x=%s y= %s", np.sum(image),
                     np.sum(clean_mask), x, y)

        if not sum(img):  # Nothing left after tail cuts
            continue

        muonringparam = muonring.fit(x, y, image * clean_mask)

        dist = np.sqrt(np.power(x - muonringparam. ring_center_x, 2)
                       + np.power(y - muonringparam.ring_center_y, 2))
        ring_dist = np.abs(dist - muonringparam.ring_radius)

        muonringparam = muonring.fit(
            x, y, img * (ring_dist < muonringparam.ring_radius * 0.4)
        )

        dist = np.sqrt(np.power(x - muonringparam.ring_center_x, 2) +
                       np.power(y - muonringparam.ring_center_y, 2))
        ring_dist = np.abs(dist - muonringparam.ring_radius)

        muonringparam = muonring.fit(
            x, y, img * (ring_dist < muonringparam.ring_radius * 0.4)
        )

        muonringparam.tel_ids = telid
        muonringparam.obs_id = event.dl0.obs_id
        muonringparam.event_id = event.dl0.event_id
        dist_mask = np.abs(dist - muonringparam.
                           ring_radius) < muonringparam.ring_radius * 0.4
        pix_im = image * dist_mask
        nom_dist = np.sqrt(np.power(muonringparam.ring_center_x,
                                    2) + np.power(muonringparam.ring_center_y, 2))

        minpix = muon_cuts['min_pix'][dict_index]  # 0.06*numpix #or 8%

        mir_rad = np.sqrt(teldes.optics.mirror_area.to("m2") / np.pi)

        # Camera containment radius -  better than nothing - guess pixel
        # diameter of 0.11, all cameras are perfectly circular   cam_rad =
        # np.sqrt(numpix*0.11/(2.*np.pi))

        if(npix_above_threshold(pix_im, tailcuts[0]) > 0.1 * minpix
           and npix_composing_ring(pix_im) > minpix
           and nom_dist < muon_cuts['CamRad'][dict_index]
           and muonringparam.ring_radius < 1.5 * u.deg
           and muonringparam.ring_radius > 1. * u.deg):
            muonringparam.ring_containment = ring_containment(
                muonringparam.ring_radius,
                muon_cuts['CamRad'][dict_index],
                muonringparam.ring_center_x,
                muonringparam.ring_center_y)

            # Guess HESS is 0.16
            # sec_rad = 0.*u.m
            # sct = False
            # if numpix == 2048 and mir_rad > 2.*u.m and mir_rad < 2.1*u.m:
            #     sec_rad = 1.*u.m
            #     sct = True
            #
            # Store muon ring parameters (passing cuts stage 1)
            # muonringlist[idx] = muonringparam

            tellist.append(telid)
            muonringlist.append(muonringparam)
            muonintensitylist.append(None)

            ctel = MuonLineIntegrate(
                mir_rad, hole_radius=muon_cuts['HoleRad'][dict_index],
                pixel_width=muon_cuts['AngPixW'][dict_index],
                sct_flag=muon_cuts['SCT'][dict_index],
                secondary_radius=muon_cuts['SecRad'][dict_index]
            )

            if image.shape[0] == muon_cuts['total_pix'][dict_index]:
                muonintensityoutput = ctel.fit_muon(muonringparam.ring_center_x,
                                                    muonringparam.ring_center_y,
                                                    muonringparam.ring_radius,
                                                    x[dist_mask], y[dist_mask],
                                                    image[dist_mask])

                muonintensityoutput.tel_ids = telid
                muonintensityoutput.obs_id = event.dl0.obs_id
                muonintensityoutput.event_id = event.dl0.event_id
                muonintensityoutput.mask = dist_mask

                idx_ring = np.nonzero(pix_im)
                muonintensityoutput.ring_completeness = ring_completeness(
                    x[idx_ring], y[idx_ring], pix_im[idx_ring],
                    muonringparam.ring_radius,
                    muonringparam.ring_center_x,
                    muonringparam.ring_center_y,
                    threshold=30,
                    bins=30)
                muonintensityoutput.ring_size = np.sum(pix_im)

                dist_ringwidth_mask = np.abs(dist - muonringparam.ring_radius
                                             ) < (muonintensityoutput.ring_width)
                pix_ringwidth_im = image * dist_ringwidth_mask
                idx_ringwidth = np.nonzero(pix_ringwidth_im)

                muonintensityoutput.ring_pix_completeness = npix_above_threshold(
                    pix_ringwidth_im[idx_ringwidth], tailcuts[0]) / len(
                    pix_im[idx_ringwidth])

                logger.debug("Tel %d Impact parameter = %s mir_rad=%s "
                             "ring_width=%s", telid,
                             muonintensityoutput.impact_parameter, mir_rad,
                             muonintensityoutput.ring_width)
                conditions = [
                    muonintensityoutput.impact_parameter * u.m <
                    muon_cuts['Impact'][dict_index][1] * mir_rad,

                    muonintensityoutput.impact_parameter
                    > muon_cuts['Impact'][dict_index][0],

                    muonintensityoutput.ring_width
                    < muon_cuts['RingWidth'][dict_index][1],

                    muonintensityoutput.ring_width
                    > muon_cuts['RingWidth'][dict_index][0]
                ]

                if all(conditions):
                    muonintensityparam = muonintensityoutput
                    idx = tellist.index(telid)
                    muonintensitylist[idx] = muonintensityparam
                    logger.debug("Muon found in tel %d,  tels in event=%d",
                                 telid, len(event.dl0.tels_with_data))
                else:
                    continue

    return muon_event_param


@deprecated('0.6')
def analyze_muon_source(source):
    """
    Generator for analyzing all the muon events

    Parameters
    ----------
    source : ctapipe.io.EventSource
        input event source

    Returns
    -------
    analyzed_muon : container
    A ctapipe event container (MuonParameter) with muon information

    """
    log.info("[FUNCTION] {}".format(__name__))

    if geom_dict is None:
        geom_dict = {}        
    numev = 0
    for event in source:  # Put a limit on number of events
        numev += 1
        analyzed_muon = analyze_muon_event(event)

        yield analyzed_muon



from ctapipe.tools.camdemo import CameraDemo
from ctapipe.tools.dump_triggers import DumpTriggersTool
from ctapipe.tools.dump_instrument import DumpInstrumentTool
from ctapipe.tools.info import info
from ctapipe.utils import get_dataset_path


def test_info():
    info(show_all=True)


def test_dump_triggers(tmpdir):
    outfile = tmpdir.join("triggers.fits")

    tool = DumpTriggersTool(
        infile=get_dataset_path("gamma_test_large.simtel.gz"),
        outfile=str(outfile)
    )

    tool.run(argv=[])

    assert outfile.exists()


def test_dump_instrument(tmpdir):
    tmpdir.chdir()

    tool = DumpInstrumentTool(
        infile=get_dataset_path("gamma_test_large.simtel.gz"),
    )

    tool.run(argv=[])

    print(tmpdir.listdir())
    assert tmpdir.join('FlashCam.camgeom.fits.gz').exists()


def test_camdemo():
    tool = CameraDemo()
    tool.num_events = 10
    tool.cleanframes = 2
    tool.display = False
    tool.run(argv=[])



from ctapipe.core import Component
from ctapipe.core.traits import Path, TraitError

from pytest import raises
import tempfile


def test_path_exists():

    class C1(Component):
        p = Path(exists=False)

    c1 = C1()
    c1.p = 'test'

    with tempfile.NamedTemporaryFile() as f:
        with raises(TraitError):
            c1.p = f.name

    class C2(Component):
        p = Path(exists=True)

    c2 = C2()

    with tempfile.TemporaryDirectory() as d:
        c2.p = d

    with tempfile.NamedTemporaryFile() as f:
        c2.p = f.name


def test_path_directory_ok():

    class C(Component):
        p = Path(exists=True, directory_ok=False)

    c = C()

    with raises(TraitError):
        c.p = 'lknasdlakndlandslknalkndslakndslkan'

    with tempfile.TemporaryDirectory() as d:
        with raises(TraitError):
            c.p = d

    with tempfile.NamedTemporaryFile() as f:
        c.p = f.name


def test_path_file_ok():

    class C(Component):
        p = Path(exists=True, file_ok=False)

    c = C()

    with raises(TraitError):
        c.p = 'lknasdlakndlandslknalkndslakndslkan'

    with tempfile.TemporaryDirectory() as d:
        c.p = d

    with tempfile.NamedTemporaryFile() as f:
        with raises(TraitError):
            c.p = f.name



from .array import get_array_layout
from .eventseeker import EventSeeker
from .eventsource import EventSource
from .eventsourcefactory import EventSourceFactory, event_source
from .hdf5tableio import HDF5TableReader, HDF5TableWriter
from .hessioeventsource import HESSIOEventSource
from .tableio import TableWriter, TableReader

__all__ = [
    'HDF5TableWriter',
    'HDF5TableReader',
    'TableWriter',
    'TableReader',
    'EventSeeker',
    'EventSourceFactory',
    'EventSource',
    'event_source'
]



import copy
from ctapipe.utils import get_dataset_path
from ctapipe.io.hessioeventsource import HESSIOEventSource


def test_hessio_file_reader():
    dataset = get_dataset_path("gamma_test.simtel.gz")
    kwargs = dict(config=None, tool=None, input_url=dataset)
    with HESSIOEventSource(**kwargs) as reader:
        assert reader.is_compatible(dataset)
        assert not reader.is_stream
        for event in reader:
            if event.count == 0:
                assert event.r0.tels_with_data == {38, 47}
            elif event.count == 1:
                assert event.r0.tels_with_data == {11, 21, 24, 26, 61, 63, 118,
                                                   119}
            else:
                break
        for event in reader:
            # Check generator has restarted from beginning
            assert event.count == 0
            break

    # test that max_events works:
    max_events = 5
    with HESSIOEventSource(**kwargs, max_events=max_events) as reader:
        count = 0
        for _ in reader:
            count += 1
        assert count == max_events

    # test that the allowed_tels mask works:
    with HESSIOEventSource(**kwargs, allowed_tels={3, 4}) as reader:
        for event in reader:
            assert event.r0.tels_with_data.issubset(reader.allowed_tels)


def test_that_event_is_not_modified_after_loop():

    dataset = get_dataset_path("gamma_test.simtel.gz")
    with HESSIOEventSource(input_url=dataset, max_events=2) as source:
        for event in source:
            last_event = copy.deepcopy(event)

        # now `event` should be identical with the deepcopy of itself from
        # inside the loop.
        # Unfortunately this does not work:
        #      assert last_event == event
        # So for the moment we just compare event ids
        assert event.r0.event_id == last_event.r0.event_id



print("import: 'ctapipe'")
import ctapipe

print("import: 'ctapipe.io'")
import ctapipe.io

print("import: 'ctapipe.core'")
import ctapipe.core

print("import: 'ctapipe.coordinates'")
import ctapipe.coordinates

print("import: 'ctapipe.instrument'")
import ctapipe.instrument

print("import: 'ctapipe.visualization'")
import ctapipe.visualization

print("import: 'ctapipe.tools'")
import ctapipe.tools

print("import: 'ctapipe.utils'")
import ctapipe.utils

print("import: 'ctapipe.reco'")
import ctapipe.reco




import numpy as np
from numpy.testing import assert_almost_equal
from copy import deepcopy

from ctapipe.image.waveform_cleaning import (NullWaveformCleaner,
                                             CHECMWaveformCleanerAverage,
                                             CHECMWaveformCleanerLocal)


def test_null_cleaner(test_event):
    telid = 11
    event = deepcopy(test_event)  # to avoid modifying the test event
    data = event.r0.tel[telid].waveform
    nsamples = data.shape[2]
    ped = event.mc.tel[telid].pedestal
    data_ped = data - np.atleast_3d(ped / nsamples)
    data_ped = np.array([data_ped[0], data_ped[0]])  # Test LG functionality

    cleaner = NullWaveformCleaner()
    cleaned = cleaner.apply(data_ped)

    assert(np.array_equal(data_ped, cleaned))


def test_checm_cleaner_average(test_event):
    telid = 11
    event = deepcopy(test_event)  # to avoid modifying the test event
    data = event.r0.tel[telid].waveform
    nsamples = data.shape[2]
    ped = event.mc.tel[telid].pedestal
    data_ped = data - np.atleast_3d(ped / nsamples)
    data_ped = np.array([data_ped[0], data_ped[0]])  # Test LG functionality

    cleaner = CHECMWaveformCleanerAverage()
    cleaned = cleaner.apply(data_ped)

    assert_almost_equal(data_ped[0, 0, 0], -2.8, 1)
    assert_almost_equal(cleaned[0, 0, 0], -6.4, 1)


def test_checm_cleaner_local(test_event):
    telid = 11
    event = deepcopy(test_event)  # to avoid modifying the test event
    data = event.r0.tel[telid].waveform
    nsamples = data.shape[2]
    ped = event.mc.tel[telid].pedestal
    data_ped = data - np.atleast_3d(ped / nsamples)
    data_ped = np.array([data_ped[0], data_ped[0]])  # Test LG functionality

    cleaner = CHECMWaveformCleanerLocal()
    cleaned = cleaner.apply(data_ped)

    assert_almost_equal(data_ped[0, 0, 0], -2.8, 1)
    assert_almost_equal(cleaned[0, 0, 0], -15.9, 1)



from pkg_resources import resource_filename
import os

import pytest
pytest.importorskip("protozfits", minversion="0.44.5")

example_file_path = resource_filename(
    'protozfits',
    os.path.join(
        'tests',
        'resources',
        'example_9evts_NectarCAM.fits.fz'
    )
)

FIRST_EVENT_NUMBER_IN_FILE = 1
ADC_SAMPLES_SHAPE = (2, 84, 60)


def test_loop_over_events():
    from ctapipe.io.nectarcameventsource import NectarCAMEventSource

    N_EVENTS = 3
    inputfile_reader = NectarCAMEventSource(
        input_url=example_file_path,
        max_events=N_EVENTS
    )

    for i, event in enumerate(inputfile_reader):
        assert event.r0.tels_with_data == [0]
        for telid in event.r0.tels_with_data:
            assert event.r0.event_id == FIRST_EVENT_NUMBER_IN_FILE + i
            assert event.r0.tel[telid].waveform.shape == ADC_SAMPLES_SHAPE

    # make sure max_events works
    assert i == N_EVENTS - 1


def test_is_compatible():
    from ctapipe.io.nectarcameventsource import NectarCAMEventSource

    assert NectarCAMEventSource.is_compatible(example_file_path)


def test_factory_for_nectarcam_file():
    from ctapipe.io.eventsourcefactory import EventSourceFactory
    from ctapipe.io.nectarcameventsource import NectarCAMEventSource

    reader = EventSourceFactory.produce(input_url=example_file_path)
    assert isinstance(reader, NectarCAMEventSource)
    assert reader.input_url == example_file_path



from matplotlib import pyplot as plt
import numpy as np
from ctapipe.visualization.mpl_array import ArrayDisplay
import astropy.units as u
from ctapipe.coordinates import GroundFrame


class NominalPlotter:
    """
    Simple plotter for drawing camera level items in the nominal system
    """

    def __init__(self, hillas_parameters, draw_axes=False, ax=None, **kwargs):
        """

        Parameters
        ----------
        instrument: dictionary
            intrument containers for this event
        telescopes: list
            List of telescopes included
        system: Coordinate system
            Coordinate system to transform coordinates into
        """
        self.axes = ax if ax is not None else plt.gca()

        self.cen_x = [i.cen_x.to(u.deg).value for i in hillas_parameters.values()]
        self.cen_y = [i.cen_y.to(u.deg).value for i in hillas_parameters.values()]

        self.centre = (0, 0)
        self.array = ArrayDisplay(telx=np.asarray(self.cen_x),
                                  tely=np.asarray(self.cen_y),
                                  tel_type=np.ones(len(self.cen_y)),
                                  axes=self.axes)

        self.hillas = hillas_parameters
        scale_fac = 57.3 * 2

        self.array.overlay_moments(hillas_parameters, (self.cen_x, self.cen_y), scale_fac,
                                   cmap="Greys", alpha=0.5, **kwargs)

        if draw_axes:
            self.array.overlay_axis(hillas_parameters, (self.cen_x, self.cen_y))

    def background_contour(self, x, y, background, **kwargs):
        """
        Draw image contours in background of the display, useful when likelihood fitting

        Parameters
        ----------
        x: ndarray
            array of image X coordinates
        y: ndarray
            array of image Y coordinates
        background: ndarray
            Array of image to use in background
        kwargs: key=value
            any style keywords to pass to matplotlib

        Returns
        -------
        None
        """

        self.axes.contour(x, y, background, **kwargs)

        # Annoyingly we need to redraw everything
        self.array = ArrayDisplay(telx=np.asarray(self.cen_x),
                                  tely=np.asarray(self.cen_y),
                                  tel_type=np.ones(len(self.cen_y)),
                                  axes=self.axes)

    def draw_array(self, coord_range=((-4, 4), (-4, 4))):
        """
        Draw the array plotter (including any overlayed elements)

        Parameters
        ----------
        coord_range: tuple
            XY range in which to draw plotter

        Returns
        -------
        None
        """

        self.array.axes.set_xlim((self.centre[0] + coord_range[0][0],
                                  coord_range[0][1] + self.centre[0]))
        self.array.axes.set_ylim((self.centre[1] + coord_range[1][0],
                                  coord_range[1][1] + self.centre[1]))

        # self.axes.tight_layout()
        # self.axes.show()

    def draw_position(self, source_x, source_y, use_centre=False, **kwargs):
        """
        Draw a marker at a position in the array plotter (for marking reconstructed
        positions etc)

        Parameters
        ----------
        source_x: float
            X position of point
        source_y: float
            Y position of point
        use_centre: bool
            Centre the plotter on this position
        kwargs: key=value
            any style keywords to pass to matplotlib

        Returns
        -------
        None
        """
        self.array.add_polygon(centroid=(source_x.to(u.deg).value,
                                         source_y.to(u.deg).value),
                               radius=0.1, nsides=3, **kwargs)
        if use_centre:
            self.centre = (source_x.value, source_y.value)



"""
Handles reading of different event/waveform containing files
"""
from abc import abstractmethod
from os.path import exists
from traitlets import Unicode, Int, Set
from ctapipe.core import Component
from ctapipe.core import Provenance

__all__ = ['EventSource', ]


class EventSource(Component):
    """
    Parent class for EventFileReaders of different sources.

    A new EventFileReader should be created for each type of event file read
    into ctapipe, e.g. sim_telarray files are read by the `HESSIOEventSource`.

    EventFileReader provides a common high-level interface for accessing event
    information from different data sources (simulation or different camera
    file formats). Creating an EventFileReader for a new
    file format ensures that data can be accessed in a common way,
    irregardless of the file format.

    EventFileReader itself is an abstract class. To use an EventFileReader you
    must use a subclass that is relevant for the file format you
    are reading (for example you must use
    `ctapipe.io.hessiofilereader.HESSIOEventSource` to read a hessio format
    file). Alternatively you can use
    `ctapipe.io.eventfilereader.EventSourceFactory` to automatically
    select the correct EventFileReader subclass for the file format you wish
    to read.

    To create an instance of an EventFileReader you must pass the traitlet
    configuration (containing the input_url) and the
    `ctapipe.core.tool.Tool`. Therefore from inside a Tool you would do:

    >>> event_source = EventSource(self.config, self)

    An example of how to use `ctapipe.core.tool.Tool` and
    `ctapipe.io.eventfilereader.EventSourceFactory` can be found in
    ctapipe/examples/calibration_pipeline.py.

    However if you are not inside a Tool, you can still create an instance and
    supply an input_url via:

    >>> event_source = EventSource( input_url="/path/to/file")

    To loop through the events in a file:

    >>> event_source = EventSource( input_url="/path/to/file")
    >>> for event in event_source:
    >>>    print(event.count)

    **NOTE**: Every time a new loop is started through the event_source, it restarts
    from the first event.

    Alternatively one can use EventFileReader in a `with` statement to ensure
    the correct cleanups are performed when you are finished with the event_source:

    >>> with EventSource( input_url="/path/to/file") as event_source:
    >>>    for event in event_source:
    >>>       print(event.count)

    **NOTE**: The "event" that is returned from the generator is a pointer.
    Any operation that progresses that instance of the generator further will
    change the data pointed to by "event". If you wish to ensure a particular
    event is kept, you should perform a `event_copy = copy.deepcopy(event)`.


    Attributes
    ----------
    input_url : str
        Path to the input event file.
    max_events : int
        Maximum number of events to loop through in generator
    metadata : dict
        A dictionary containing the metadata of the file. This could include:
        * is_simulation (bool indicating if the file contains simulated events)
        * Telescope:Camera names (list if file contains multiple)
        * Information in the file header
        * Observation ID
    """

    input_url = Unicode(
        '',
        help='Path to the input file containing events.'
    ).tag(config=True)
    max_events = Int(
        None,
        allow_none=True,
        help='Maximum number of events that will be read from the file'
    ).tag(config=True)

    allowed_tels = Set(
        help=('list of allowed tel_ids, others will be ignored. '
              'If left empty, all telescopes in the input stream '
              'will be included')
    ).tag(config=True)


    def __init__(self, config=None, tool=None, **kwargs):
        """
        Class to handle generic input files. Enables obtaining the "source"
        generator, regardless of the type of file (either hessio or camera
        file).

        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        kwargs
        """
        super().__init__(config=config, parent=tool, **kwargs)

        self.metadata = dict(is_simulation=False)

        if not exists(self.input_url):
            raise FileNotFoundError("file path does not exist: '{}'"
                                    .format(self.input_url))
        self.log.info("INPUT PATH = {}".format(self.input_url))

        if self.max_events:
            self.log.info("Max events being read = {}".format(self.max_events))

        Provenance().add_input_file(self.input_url, role='dl0.sub.evt')

    @staticmethod
    @abstractmethod
    def is_compatible(file_path):
        """
        Abstract method to be defined in child class.

        Perform a set of checks to see if the input file is compatible
        with this file event_source.

        Parameters
        ----------
        file_path : str
            File path to the event file.

        Returns
        -------
        compatible : bool
            True if file is compatible, False if it is incompatible
        """

    @property
    def is_stream(self):
        """
        Bool indicating if input is a stream. If it is then it is incompatible
        with `ctapipe.io.eventseeker.EventSeeker`.

        TODO: Define a method to detect if it is a stream

        Returns
        -------
        bool
            If True, then input is a stream.
        """
        return False

    @abstractmethod
    def _generator(self):
        """
        Abstract method to be defined in child class.

        Generator where the filling of the `ctapipe.io.containers` occurs.

        Returns
        -------
        generator
        """

    def __iter__(self):
        """
        Generator that iterates through `_generator`, but keeps track of
        `self.max_events`.

        Returns
        -------
        generator
        """
        for event in self._generator():
            yield event
            if self.max_events and event.count >= self.max_events - 1:
                break

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



"""
Handles seeking to a particular event in a
`ctapipe.io.eventfilereader.EventFileReader`
"""
from copy import deepcopy
from ctapipe.core import Component

__all__ = ['EventSeeker', ]


class EventSeeker(Component):
    """
    Provides the functionality to seek through a
    `ctapipe.io.eventfilereader.EventSource` to find a particular event.

    By default, this will loop through events from the start of the file
    (unless the requested event is the same as the previous requested event,
    or occurs later in the file). However if the
    `ctapipe.io.eventfilereader.EventSource` has defined a `__getitem__`
    method itself, then it will use that method, thereby taking advantage of
    the random event access some file formats provide.

    To create an instance of an EventSeeker you must provide it a sub-class of
    `ctapipe.io.eventfilereader.EventSource` (such as
    `ctapipe.io.hessiofilereader.HessioFileReader`), which will be used to
    loop through the file and provide the event container, filled with the
    event information using the methods defined in the event_source for that
    file format.

    To obtain a particular event in a hessio file:

    >>> from ctapipe.io.hessioeventsource import HESSIOEventSource
    >>> event_source = HESSIOEventSource(input_path="/path/to/file")
    >>> seeker = EventSeeker(event_source=event_source)
    >>> event = seeker[2]
    >>> print(event.count)

    To obtain a particular event in a hessio file from its event_id:

    >>> from ctapipe.io.hessioeventsource import HESSIOEventSource
    >>> event_source = HESSIOEventSource(input_path="/path/to/file")
    >>> seeker = EventSeeker(event_source=event_source)
    >>> event = seeker["101"]
    >>> print(event.count)

    **NOTE**: Event_index refers to the number associated to the event
    assigned by ctapipe (`event.count`), based on the order the events are
    read from the file.
    Whereas the event_id refers to the ID attatched to the event from the
    external source of the file (software or camera or CTA array).

    To obtain a slice of events in a hessio file:

    >>> from ctapipe.io.hessioeventsource import HESSIOEventSource
    >>> event_source = HESSIOEventSource(input_path="/path/to/file")
    >>> seeker = EventSeeker(event_source=event_source)
    >>> event_list = seeker[3:6]
    >>> print([event.count for event in event_list])

    To obtain a list of events in a hessio file:

    >>> from ctapipe.io.hessioeventsource import HESSIOEventSource
    >>> event_source = HESSIOEventSource(input_path="/path/to/file")
    >>> seeker = EventSeeker(event_source)
    >>> event_indicis = [2, 6, 8]
    >>> event_list = seeker[event_indicis]
    >>> print([event.count for event in event_list])
    """

    def __init__(self, reader, config=None, tool=None, **kwargs):
        """
        Class to handle generic input files. Enables obtaining the "source"
        generator, regardless of the type of file (either hessio or camera
        file).

        Parameters
        ----------
        reader : `ctapipe.io.eventfilereader.EventSource`
            A subclass of `ctapipe.io.eventfilereader.EventFileReader` that
            defines how the event container is filled for a particular file
            format
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        kwargs
        """
        super().__init__(config=config, parent=tool, **kwargs)

        if reader.is_stream:
            raise IOError("Reader is not compatible as input to the "
                          "event_source is a stream (seeking not possible)")

        self._reader = reader

        self._num_events = None
        self._source = self._reader.__iter__()
        self._current_event = None
        self._has_fast_seek = False  # By default seeking iterates through
        self._getevent_warn = True

    def _reset(self):
        """
        Recreate the generator so it starts from the beginning
        """
        self._source = self._reader.__iter__()
        self._current_event = None

    def __iter__(self):
        # Always reset generator when starting a new iteration
        self._reset()
        for event in self._source:
            self._current_event = event
            yield event

    def __getitem__(self, item):
        """
        Obtain a particular event

        Parameters
        ----------
        item : int or str
            If `item` is an int, then this is the event_index for the event
            obtained. If `item` is a str, then this is the event_id for the
            event obtained.

        Returns
        -------
        event : ctapipe.io.container
            The event container filled with the requested event's information

        """

        # Handling of different input types (int, string, slice, list)
        current = None
        use_event_id = False
        if isinstance(item, int):
            if self._current_event:
                current = self._current_event.count
            if item < 0:
                item = len(self) + item
                if item < 0 or item >= len(self):
                    msg = ("Event index {} out of range [0, {}]"
                           .format(item, len(self)))
                    raise IndexError(msg)
        elif isinstance(item, str):
            item = int(item)
            use_event_id = True
            if self._current_event:
                current = self._current_event.r0.event_id
        elif isinstance(item, slice):
            it = range(item.start or 0, item.stop or len(self), item.step or 1)
            events = [self[i] for i in it]
            return events
        elif isinstance(item, list):
            events = [self[i] for i in item]
            return events
        else:
            raise TypeError("{} indexing is not supported".format(type(item)))

        # Return a copy of the current event if we have already reached it
        if current is not None and item == current:
            return deepcopy(self._current_event)

        # If requested event is less than the current event position: reset
        if current is not None and item < current:
            self._reset()

        # Check we are within max_events range
        max_events = self._reader.max_events
        if not use_event_id and max_events and item >= max_events:
            msg = ("Event index {} outside of specified max_events {}"
                   .format(item, max_events))
            raise IndexError(msg)

        try:
            if not use_event_id:
                event = self._reader._get_event_by_index(item)
            else:
                event = self._reader._get_event_by_id(item)
        except AttributeError:
            if self._getevent_warn:
                self.log.warning("Seeking to event by looping through "
                                 "events... (potentially long process)")
                self._getevent_warn = False
            if not use_event_id:
                event = self._get_event_by_index(item)
            else:
                event = self._get_event_by_id(item)

        self._current_event = event
        return deepcopy(event)

    def _get_event_by_index(self, index):
        """
        Method for extracting a particular event by looping through events
        until it finds the requested event index.
        If a file format allows random event access, then is can define its
        own `get_event_by_index` method in its
        `ctapipe.io.eventfilereader.EventSource` to allow this class to
        utilise that method instead.

        Parameters
        ----------
        index : int
            The event_index to seek.

        Returns
        -------
        event : ctapipe.io.container
            The event container filled with the requested event's information

        """
        for event in self._source:
            if event.count == index:
                return event
        raise IndexError("Event index {} not found in file".format(index))

    def _get_event_by_id(self, event_id):
        """
        Method for extracting a particular event by looping through events
        until it finds the requested event id.
        If a file format allows random event access, then is can define its
        own `get_event_by_id` method in its
        `ctapipe.io.eventfilereader.EventSource` to allow this class to
        utilise that method instead.

        Parameters
        ----------
        event_id : int
            The event_id to seek.

        Returns
        -------
        event : ctapipe.io.container
            The event container filled with the requested event's information

        """
        for event in self:  # Event Ids may not be in order
            if event.r0.event_id == event_id:
                return event
        raise IndexError("Event id {} not found in file".format(event_id))

    def __len__(self):
        """
        Method for getting number of events in file. By default this is
        obtained by looping through the file and counting the events. If a
        file format has a more efficient method of supplying this information,
        the `ctapipe.io.eventfilereader.EventSource` for that file format
        can define its own `__len__` method, which this class will then
        use instead.

        Returns
        -------
        self._num_events : int
            Number of events in the file
        """
        # Only need to calculate once
        if not self._num_events:
            try:
                count = len(self._reader)
            except TypeError:
                self.log.warning("Obtaining length of file by looping through "
                                 "all events... (potentially long process)")
                count = 0
                for _ in self:
                    count += 1
            self._num_events = count
        return self._num_events



"""
Example tool, displaying fake events in a camera.

the animation should remain interactive, so try zooming in when it is
running.
"""

import matplotlib.pylab as plt
import numpy as np
from astropy import units as u
from matplotlib.animation import FuncAnimation

from ctapipe.core import Tool, traits
from ctapipe.image import toymodel, tailcuts_clean, dilate
from ctapipe.instrument import TelescopeDescription, CameraGeometry, \
    OpticsDescription
from ctapipe.visualization import CameraDisplay


class CameraDemo(Tool):
    name = u"ctapipe-camdemo"
    description = "Display fake events in a demo camera"

    delay = traits.Int(50, help="Frame delay in ms", min=20).tag(config=True)
    cleanframes = traits.Int(100, help="Number of frames between turning on "
                                       "cleaning", min=0).tag(config=True)
    autoscale = traits.Bool(False, help='scale each frame to max if '
                                        'True').tag(config=True)
    blit = traits.Bool(False, help='use blit operation to draw on screen ('
                                   'much faster but may cause some draw '
                                   'artifacts)').tag(config=True)
    camera = traits.CaselessStrEnum(
        CameraGeometry.get_known_camera_names(),
        default_value='NectarCam',
        help='Name of camera to display').tag(config=True)

    optics = traits.CaselessStrEnum(
        OpticsDescription.get_known_optics_names(),
        default_value='MST',
        help='Telescope optics description name'
    ).tag(config=True)

    num_events = traits.Int(0, help='events to show before exiting (0 for '
                                    'unlimited)').tag(config=True)

    display = traits.Bool(True, "enable or disable display (for "
                                "testing)").tag(config=True)

    aliases = traits.Dict({
        'delay': 'CameraDemo.delay',
        'cleanframes': 'CameraDemo.cleanframes',
        'autoscale': 'CameraDemo.autoscale',
        'blit': 'CameraDemo.blit',
        'camera': 'CameraDemo.camera',
        'optics': 'CameraDemo.optics',
        'num-events': 'CameraDemo.num_events'
    })

    def __init__(self):
        super().__init__()
        self._counter = 0
        self.imclean = False

    def start(self):
        self.log.info("Starting CameraDisplay for {}".format(self.camera))
        self._display_camera_animation()

    def _display_camera_animation(self):
        # plt.style.use("ggplot")
        fig = plt.figure(num="ctapipe Camera Demo", figsize=(7, 7))
        ax = plt.subplot(111)

        # load the camera
        tel = TelescopeDescription.from_name(optics_name=self.optics,
                                             camera_name=self.camera)
        geom = tel.camera

        # poor-man's coordinate transform from telscope to camera frame (it's
        # better to use ctapipe.coordiantes when they are stable)
        scale = tel.optics.equivalent_focal_length.to(geom.pix_x.unit).value
        fov = np.deg2rad(4.0)
        maxwid = np.deg2rad(0.01)
        maxlen = np.deg2rad(0.03)

        disp = CameraDisplay(
            geom, ax=ax, autoupdate=True,
            title="{}, f={}".format(tel, tel.optics.equivalent_focal_length)
        )
        disp.cmap = plt.cm.terrain

        def update(frame):

            self.log.debug("Frame=", frame)
            centroid = np.random.uniform(-fov, fov, size=2) * scale
            width = np.random.uniform(0, maxwid) * scale
            length = np.random.uniform(0, maxlen) * scale + width
            angle = np.random.uniform(0, 360)
            intens = np.random.exponential(2) * 50
            model = toymodel.generate_2d_shower_model(centroid=centroid,
                                                      width=width,
                                                      length=length,
                                                      psi=angle * u.deg)
            image, sig, bg = toymodel.make_toymodel_shower_image(geom,
                                                                 model.pdf,
                                                                 intensity=intens,
                                                                 nsb_level_pe=5000)

            # alternate between cleaned and raw images
            if self._counter == self.cleanframes:
                plt.suptitle("Image Cleaning ON")
                self.imclean = True
            if self._counter == self.cleanframes * 2:
                plt.suptitle("Image Cleaning OFF")
                self.imclean = False
                self._counter = 0

            if self.imclean:
                cleanmask = tailcuts_clean(geom, image / 80.0)
                for ii in range(3):
                    dilate(geom, cleanmask)
                image[cleanmask == 0] = 0  # zero noise pixels

            self.log.debug("count = {}, image sum={} max={}"
                           .format(self._counter, image.sum(), image.max()))
            disp.image = image

            if self.autoscale:
                disp.set_limits_percent(95)
            else:
                disp.set_limits_minmax(-100, 4000)

            disp.axes.figure.canvas.draw()
            self._counter += 1
            return [ax, ]

        frames = None if self.num_events == 0 else self.num_events
        repeat = True if self.num_events == 0 else False

        self.log.info("Running for {} frames".format(frames))
        self.anim = FuncAnimation(fig, update,
                                  interval=self.delay,
                                  frames=frames,
                                  repeat=repeat,
                                  blit=self.blit)

        if self.display:
            plt.show()


def main():
    app = CameraDemo()
    app.run()


if __name__ == '__main__':
    main()



import numpy as np
from astropy import units as u
from ctapipe.instrument import CameraGeometry
from ctapipe.instrument.camera import (
    _find_neighbor_pixels,
    _get_min_pixel_seperation,
)
import pytest

cam_ids = CameraGeometry.get_known_camera_names()


def test_construct():
    x = np.linspace(-10, 10, 100)
    y = np.linspace(-10, 10, 100)
    geom = CameraGeometry(cam_id=0, pix_id=np.arange(100),
                          pix_x=x * u.m, pix_y=y * u.m,
                          pix_area=x * u.m**2,
                          pix_type='rectangular',
                          pix_rotation="10d",
                          cam_rotation="12d")

    assert geom.cam_id == 0
    assert geom.pix_area is not None
    assert (geom.pix_rotation.deg - 10) < 1e-5
    assert (geom.cam_rotation.deg - 10) < 1e-5


def test_known_camera_names():
    cams = CameraGeometry.get_known_camera_names()
    assert len(cams) > 4
    assert 'FlashCam' in cams
    assert 'NectarCam' in cams

    for cam in cams:
        geom = CameraGeometry.from_name(cam)
        geom.info()


def test_make_rectangular_camera_geometry():
    geom = CameraGeometry.make_rectangular()
    assert geom.pix_x.shape == geom.pix_y.shape


def test_load_hess_camera():
    geom = CameraGeometry.from_name("LSTCam")
    assert len(geom.pix_x) == 1855


def test_guess_camera():
    px = np.linspace(-10, 10, 11328) * u.m
    py = np.linspace(-10, 10, 11328) * u.m
    geom = CameraGeometry.guess(px, py, 0 * u.m)
    assert geom.pix_type.startswith('rect')


def test_get_min_pixel_seperation():
    x, y = np.meshgrid(np.linspace(-5, 5, 5), np.linspace(-5, 5, 5))
    pixsep = _get_min_pixel_seperation(x.ravel(), y.ravel())
    assert pixsep == 2.5


def test_find_neighbor_pixels():
    x, y = np.meshgrid(np.linspace(-5, 5, 5), np.linspace(-5, 5, 5))
    neigh = _find_neighbor_pixels(x.ravel(), y.ravel(), rad=3.1)
    assert set(neigh[11]) == set([16, 6, 10, 12])


@pytest.mark.parametrize("cam_id", cam_ids)
def test_neighbor_pixels(cam_id):
    """
    test if each camera has a reasonable number of neighbor pixels (4 for
    rectangular, and 6 for hexagonal.  Other than edge pixels, the majority
    should have the same value
    """

    geom = CameraGeometry.from_name(cam_id)
    n_pix = len(geom.pix_id)
    n_neighbors = [len(x) for x in geom.neighbors]

    if geom.pix_type.startswith('hex'):
        assert n_neighbors.count(6) > 0.5 * n_pix
        assert n_neighbors.count(6) > n_neighbors.count(4)

    if geom.pix_type.startswith('rect'):
        assert n_neighbors.count(4) > 0.5 * n_pix
        assert n_neighbors.count(5) == 0
        assert n_neighbors.count(6) == 0

    assert n_neighbors.count(1) == 0  # no pixel should have a single neighbor


def test_to_and_from_table():
    geom = CameraGeometry.from_name("LSTCam")
    tab = geom.to_table()
    geom2 = geom.from_table(tab)

    assert geom.cam_id == geom2.cam_id
    assert (geom.pix_x == geom2.pix_x).all()
    assert (geom.pix_y == geom2.pix_y).all()
    assert (geom.pix_area == geom2.pix_area).all()
    assert geom.pix_type == geom2.pix_type


def test_write_read(tmpdir):

    filename = str(tmpdir.join('testcamera.fits.gz'))

    geom = CameraGeometry.from_name("LSTCam")
    geom.to_table().write(filename, overwrite=True)
    geom2 = geom.from_table(filename)

    assert geom.cam_id == geom2.cam_id
    assert (geom.pix_x == geom2.pix_x).all()
    assert (geom.pix_y == geom2.pix_y).all()
    assert (geom.pix_area == geom2.pix_area).all()
    assert geom.pix_type == geom2.pix_type


def test_precal_neighbors():
    geom = CameraGeometry(cam_id="TestCam",
                          pix_id=np.arange(3),
                          pix_x=np.arange(3) * u.deg,
                          pix_y=np.arange(3) * u.deg,
                          pix_area=np.ones(3) * u.deg**2,
                          neighbors=[
                              [1, ], [0, 2], [1, ]
                          ],
                          pix_type='rectangular',
                          pix_rotation="0deg",
                          cam_rotation="0deg")

    neigh = geom.neighbors
    assert len(neigh) == len(geom.pix_x)

    nmat = geom.neighbor_matrix
    assert nmat.shape == (len(geom.pix_x), len(geom.pix_x))


def test_slicing():
    geom = CameraGeometry.from_name("NectarCam")
    sliced1 = geom[100:200]

    assert len(sliced1.pix_x) == 100
    assert len(sliced1.pix_y) == 100
    assert len(sliced1.pix_area) == 100
    assert len(sliced1.pix_id) == 100

    sliced2 = geom[[5, 7, 8, 9, 10]]
    assert sliced2.pix_id[0] == 5
    assert sliced2.pix_id[1] == 7
    assert len(sliced2.pix_x) == 5



"""
This class is largely created as a more versatile alternative to the Linear ND
interpolator, allowing the user to pass the interpolator an array of classes and then
perform interpolation between the results of a member function of that class. Currently
the name of the target interpolation function is passed in a string at initialisation.

TODO:
- Figure out what to do when out of bounds of interpolation range
- Create function to append points to the interpolator

"""

import numpy as np
from scipy.spatial import Delaunay


class UnstructuredInterpolator:
    """
    This class performs linear interpolation between an unstructured set of data
    points. However the class is expanded such that it can interpolate between the
    values returned from class member functions. The primary use case of this being to
    interpolate between the predictions of a set of machine learning algorithms or
    regular grid interpolators.

    In the case that a numpy array is passed as the interpolation values this class will
    behave exactly the same as the scipy LinearNDInterpolator
    """
    def __init__(self, interpolation_points, function_name=None):
        """
        Parameters
        ----------
        interpolation_points: dict
            Dictionary of interpolation points (stored as key) and values
        function_name: str
            Name of class member function to call in the case we are interpolating
            between class predictions, for numpy arrays leave blank
        """

        self.keys = np.array(list(interpolation_points.keys()))
        self.values = np.array(list(interpolation_points.values()))

        self._num_dimensions = len(self.keys[0])

        # create an object with triangulation
        self._tri = Delaunay(self.keys)
        self._function_name = function_name

        # OK this code is horrid and will need fixing
        self._numpy_input = isinstance(self.values[0], np.ndarray) or \
                            issubclass(type(self.values[0]), np.float) or \
                            issubclass(type(self.values[0]), np.int)

        if self._numpy_input is False and function_name is None:
            self._function_name = "__call__"

        return None

    def __call__(self, points, eval_points=None):

        if self._numpy_input is False and np.all(eval_points is None):
            raise ValueError("Non numpy object provided without with emtpy eval_points")

        # Convert to a numpy array here incase we get a list
        points = np.array(points)

        if len(points.shape) == 1:
            points = np.array([points])

        # find simplexes that contain interpolated points
        s = self._tri.find_simplex(points)
        # get the vertices for each simplex
        v = self._tri.vertices[s]
        # get transform matrices for each simplex (see explanation bellow)
        m = self._tri.transform[s]

        # Here comes some serious numpy magic, it could be done with a loop but would
        # be pretty inefficient I had to rip this from stack overflow - RDP
        # For each interpolated point, take the the transform matrix and multiply it by
        # the vector p-r, where r=m[:,n,:] is one of the simplex vertices to which
        # the matrix m is related to
        b = np.einsum('ijk,ik->ij', m[:, :self._num_dimensions, :self._num_dimensions],
                      points - m[:, self._num_dimensions, :])

        # Use the above array to get the weights for the vertices; `b` contains an
        # n-dimensional vector with weights for all but the last vertices of the simplex
        # (note that for n-D grid, each simplex consists of n+1 vertices);
        # the remaining weight for the last vertex can be copmuted from
        # the condition that sum of weights must be equal to 1
        w = np.c_[b, 1 - b.sum(axis=1)]

        if self._numpy_input:
            selected_points = self.values[v]
        else:
            selected_points = self._call_class_function(v, eval_points)

        # Multiply point values by weight
        p_values = np.einsum('ij...,ij...->i...', selected_points, w)
        return p_values

    def _call_class_function(self, point_num, eval_points):
        """
        Function to loop over class function and return array of outputs

        Parameters
        ----------
        point_num: int
            Index of class position in values list
        eval_points: ndarray
            Inputs used to evaluate class member function

        Returns
        -------
        ndarray: output from member function
        """

        outputs = list()
        shape = point_num.shape
        for pt in point_num.ravel():
            cls = self.values[pt]
            cls_function = getattr(cls, self._function_name)
            outputs.append(cls_function(eval_points))

        outputs = np.array(outputs)
        new_shape = (*shape, *outputs.shape[1:])

        outputs = outputs.reshape(new_shape)

        return outputs
